"""
Tools for interacting with realtime execution contexts.
"""

import asyncio
import concurrent.futures
import enum
import logging
import threading
import warnings
from collections.abc import Sequence as SequenceABC
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    SupportsInt,
    Type,
    Union,
    cast,
)

from uqbar.objects import new

from ..assets.synthdefs import system_synthdefs
from ..enums import BootStatus, CalculationRate
from ..exceptions import (
    OwnedServerShutdown,
    ServerCannotBoot,
    ServerOffline,
    ServerOnline,
    TooManyClients,
    UnownedServerShutdown,
)
from ..osc import (
    AsyncOscProtocol,
    HealthCheck,
    OscBundle,
    OscMessage,
    OscProtocol,
    OscProtocolOffline,
    ThreadedOscProtocol,
)
from ..scsynth import AsyncProcessProtocol, Options, SyncProcessProtocol
from ..typing import SupportsOsc
from ..ugens import SynthDef
from .core import (
    Buffer,
    Bus,
    Context,
    ContextObject,
    Group,
    InvalidCalculationRate,
    Node,
    Synth,
)
from .requests import (
    DumpTree,
    GetBuffer,
    GetBufferRange,
    GetControlBus,
    GetControlBusRange,
    GetSynthControl,
    GetSynthControlRange,
    QueryBuffer,
    QueryNode,
    QueryStatus,
    QueryTree,
    QueryVersion,
    Quit,
    Sync,
    ToggleNotifications,
)
from .responses import (
    BufferInfo,
    DoneInfo,
    FailInfo,
    GetBufferInfo,
    GetBufferRangeInfo,
    GetControlBusInfo,
    GetControlBusRangeInfo,
    GetNodeControlInfo,
    GetNodeControlRangeInfo,
    NodeInfo,
    QueryTreeGroup,
    QueryTreeInfo,
    StatusInfo,
    VersionInfo,
)

if TYPE_CHECKING:
    from ..realtime.shm import ServerSHM

logger = logging.getLogger(__name__)


class FailWarning(Warning):
    pass


warnings.simplefilter("always", FailWarning)


DEFAULT_HEALTHCHECK = HealthCheck(
    active=False,
    backoff_factor=1.5,
    max_attempts=5,
    request_pattern=["/status"],
    response_pattern=["/status.reply"],
    timeout=1.0,
)


class ServerLifecycleEvent(enum.Enum):
    BOOTING = enum.auto()
    PROCESS_BOOTED = enum.auto()
    CONNECTING = enum.auto()
    OSC_CONNECTED = enum.auto()
    CONNECTED = enum.auto()
    BOOTED = enum.auto()
    OSC_PANICKED = enum.auto()
    PROCESS_PANICKED = enum.auto()
    QUITTING = enum.auto()
    DISCONNECTING = enum.auto()
    OSC_DISCONNECTED = enum.auto()
    DISCONNECTED = enum.auto()
    PROCESS_QUIT = enum.auto()
    QUIT = enum.auto()


class ServerShutdownEvent(enum.Enum):
    QUIT = enum.auto()
    DISCONNECT = enum.auto()
    OSC_PANIC = enum.auto()
    PROCESS_PANIC = enum.auto()
    TOO_MANY_CLIENTS = enum.auto()


class BaseServer(Context):
    """
    Base class for realtime execution contexts.

    :param options: The context's options.
    :param kwargs: Keyword arguments for options.
    """

    ### CLASS VARIABLES ###

    _contexts: Set["BaseServer"] = set()

    ### INITIALIZER ###

    def __init__(
        self,
        options: Optional[Options],
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(options, name=name, **kwargs)
        self._boot_status = BootStatus.OFFLINE
        self._buffers: Set[int] = set()
        self._is_owner = False
        self._latency = 0.1
        self._lifecycle_event_callbacks: Dict[ServerLifecycleEvent, List[Callable]] = {}
        self._maximum_logins = 1
        self._node_active: Dict[int, bool] = {}
        self._node_children: Dict[int, List[int]] = {}
        self._node_parents: Dict[int, int] = {}
        self._shm: Optional["ServerSHM"] = None
        self._status: Optional[StatusInfo] = None

    ### SPECIAL METHODS ###

    def __contains__(self, object_: ContextObject) -> bool:
        if object_.context is not self:
            return False
        if self.boot_status != BootStatus.ONLINE:
            return False
        if isinstance(object_, Node) and (
            object_.id_ in self._node_parents or object_.id_ == 0
        ):
            return True
        if isinstance(object_, Buffer) and object_.id_ in self._buffers:
            return True
        if isinstance(object_, Bus):
            if object_.calculation_rate == CalculationRate.AUDIO and (
                object_.id_ < self.options.control_bus_channel_count
            ):
                return True
            elif object_.calculation_rate == CalculationRate.CONTROL and (
                object_.id_ < self.options.audio_bus_channel_count
            ):
                return True
        return False

    ### PRIVATE METHODS ###

    def _add_node_to_children(
        self, id_: int, parent_id: int, previous_id: int, next_id: int
    ) -> None:
        self._node_parents[id_] = parent_id
        children = self._node_children[parent_id]
        if previous_id == -1:
            children.insert(0, id_)
        elif next_id == -1:
            children.append(id_)
        elif previous_id in children:
            children.insert(children.index(previous_id) + 1, id_)
        elif next_id in children:
            children.insert(children.index(next_id), id_)

    def _free_id(
        self,
        type_: Type[ContextObject],
        id_: int,
        calculation_rate: Optional[CalculationRate] = None,
    ) -> None:
        self._get_allocator(type_, calculation_rate).free(id_)

    def _handle_done_b_alloc(self, message: OscMessage) -> None:
        with self._lock:
            self._buffers.add(message.contents[1])

    def _handle_done_b_alloc_read(self, message: OscMessage) -> None:
        with self._lock:
            self._buffers.add(message.contents[1])

    def _handle_done_b_alloc_read_channel(self, message: OscMessage) -> None:
        with self._lock:
            self._buffers.add(message.contents[1])

    def _handle_done_b_free(self, message: OscMessage) -> None:
        with self._lock:
            if message.contents[1] in self._buffers:
                self._buffers.remove(message.contents[1])
            self._free_id(Buffer, message.contents[1])

    def _handle_fail(self, message: OscMessage) -> None:
        warnings.warn(" ".join(str(x) for x in message.contents), FailWarning)

    def _handle_n_end(self, message: OscMessage) -> None:
        with self._lock:
            id_, parent_id, *_ = message.contents
            if parent_id == -1:
                parent_id = self._node_parents.get(id_)
            if parent_id is not None:
                self._remove_node_from_children(id_, parent_id)
            self._free_id(Node, id_)
            self._node_active.pop(id_, None)
            self._node_children.pop(id_, None)
            self._node_parents.pop(id_, None)

    def _handle_n_go(self, message: OscMessage) -> None:
        with self._lock:
            id_, parent_id, previous_id, next_id, is_group, *_ = message.contents
            self._node_parents[id_] = parent_id
            self._node_active[id_] = True
            if is_group:
                self._node_children[id_] = []
            self._add_node_to_children(id_, parent_id, previous_id, next_id)

    def _handle_n_move(self, message: OscMessage) -> None:
        with self._lock:
            id_, parent_id, previous_id, next_id, *_ = message.contents
            old_parent_id = self._node_parents[id_]
            self._remove_node_from_children(id_, old_parent_id)
            self._add_node_to_children(id_, parent_id, previous_id, next_id)

    def _handle_n_off(self, message: OscMessage) -> None:
        with self._lock:
            self._node_active[message.contents[0]] = False

    def _handle_n_on(self, message: OscMessage) -> None:
        with self._lock:
            self._node_active[message.contents[0]] = True

    def _handle_status_reply(self, message: OscMessage):
        with self._lock:
            self._status = cast(StatusInfo, StatusInfo.from_osc(message))

    def _log_prefix(self) -> str:
        return f"[{self._options.ip_address}:{self._options.port}/{self.name or hex(id(self))}] "

    def _remove_node_from_children(self, id_: int, parent_id: int) -> None:
        if not (children := self._node_children.get(parent_id, [])):
            return
        try:
            children.pop(children.index(id_))
        except ValueError:
            pass

    def _resolve_node(self, node: Union[Node, SupportsInt, None]) -> int:
        if node is None:
            return self._client_id + 1
        return int(node)

    def _setup_osc_callbacks(self, osc_protocol: OscProtocol) -> None:
        for pattern, procedure in [
            (["/done", "/b_alloc"], self._handle_done_b_alloc),
            (["/done", "/b_allocRead"], self._handle_done_b_alloc_read),
            (["/done", "/b_allocReadChannel"], self._handle_done_b_alloc_read_channel),
            (["/done", "/b_free"], self._handle_done_b_free),
            (["/fail"], self._handle_fail),
            (["/n_end"], self._handle_n_end),
            (["/n_go"], self._handle_n_go),
            (["/n_move"], self._handle_n_move),
            (["/n_off"], self._handle_n_off),
            (["/n_on"], self._handle_n_on),
            (["/status.reply"], self._handle_status_reply),
        ]:
            osc_protocol.register(pattern=pattern, procedure=procedure)

    def _setup_shm(self) -> None:
        try:
            from .shm import ServerSHM

            self._shm = ServerSHM(
                self._options.port, self._options.control_bus_channel_count
            )
        except (ImportError, ModuleNotFoundError):
            pass

    def _setup_system(self) -> None:
        self._node_children[0] = []
        with self.at():
            for i in range(self._maximum_logins):
                self.add_group(permanent=True, add_action="ADD_TO_TAIL", target_node=0)
        for name in dir(system_synthdefs):
            synthdef = getattr(system_synthdefs, name)
            if isinstance(synthdef, SynthDef):
                with self.at():
                    self.add_synthdefs(synthdef)

    def _teardown_shm(self) -> None:
        self._shm = None

    def _teardown_state(self) -> None:
        self._node_active.clear()
        self._node_children.clear()
        self._node_parents.clear()
        self._buffers.clear()

    def _validate_can_request(self) -> None:
        if self._boot_status not in (BootStatus.BOOTING, BootStatus.ONLINE):
            raise ServerOffline
        pass  # Otherwise always OK to request in RT

    def _validate_moment_timestamp(self, seconds: Optional[float]) -> None:
        pass  # Floats and None are OK in RT

    ### PUBLIC METHODS ###

    def on(
        self,
        event: Union[ServerLifecycleEvent, Iterable[ServerLifecycleEvent]],
        callback: Callable[[ServerLifecycleEvent], None],
    ) -> None:
        if isinstance(event, ServerLifecycleEvent):
            events_ = [event]
        else:
            events_ = list(set(event))
        for event_ in events_:
            if callback not in (
                callbacks := self._lifecycle_event_callbacks.setdefault(event_, [])
            ):
                callbacks.append(callback)

    def send(
        self, message: Union[OscMessage, OscBundle, SupportsOsc, SequenceABC, str]
    ) -> None:
        """
        Send a message to the execution context.

        :param message: The message to send.
        """
        if self._boot_status == BootStatus.OFFLINE:
            raise ServerOffline
        osc_protocol: OscProtocol = getattr(self, "_osc_protocol")
        osc_protocol.send(
            message.to_osc() if isinstance(message, SupportsOsc) else message
        )

    def set_latency(self, latency: float) -> None:
        """
        Set the context's latency.

        :param latency: The latency in seconds.
        """
        self._latency = float(latency)

    ### PUBLIC PROPERTIES ###

    @property
    def default_group(self) -> Group:
        """
        Get the server's default group.
        """
        return Group(context=self, id_=self._client_id + 1)

    @property
    def is_owner(self) -> bool:
        """
        Get the server's ownership flag.
        """
        return self._is_owner

    @property
    def status(self) -> Optional[StatusInfo]:
        """
        Get the server's last received status.
        """
        return self._status


class Server(BaseServer):
    """
    A realtime execution context with :py:mod:`threading`-based OSC and process protocols.

    :param options: The context's options.
    :param kwargs: Keyword arguments for options.
    """

    ### INITIALIZER ###

    def __init__(
        self,
        options: Optional[Options] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            options=options,
            **kwargs,
        )
        self._boot_future: concurrent.futures.Future[bool] = concurrent.futures.Future()
        self._exit_future: concurrent.futures.Future[bool] = concurrent.futures.Future()
        self._shutdown_future: concurrent.futures.Future[ServerShutdownEvent] = (
            concurrent.futures.Future()
        )
        self._osc_protocol = ThreadedOscProtocol(
            name=name,
            on_panic_callback=lambda: self._shutdown_future.set_result(
                ServerShutdownEvent.OSC_PANIC
            ),
        )
        self._process_protocol = SyncProcessProtocol(
            name=name,
            on_panic_callback=lambda: self._shutdown_future.set_result(
                ServerShutdownEvent.PROCESS_PANIC
            ),
        )
        self._setup_osc_callbacks(self._osc_protocol)

    ### PRIVATE METHODS ###

    def _lifecycle(self, owned=True) -> None:
        log_prefix = self._log_prefix()
        logger.info(log_prefix + "booting ...")
        if owned:
            self._on_lifecycle_event(ServerLifecycleEvent.BOOTING)
            try:
                self._process_protocol.boot(self._options)
            except ServerCannotBoot:
                self._on_lifecycle_event(ServerLifecycleEvent.PROCESS_PANICKED)
                self._boot_status = BootStatus.OFFLINE
                self._boot_future.set_result(False)
                self._exit_future.set_result(False)
                return
            self._is_owner = True
            self._setup_shm()
            self._on_lifecycle_event(ServerLifecycleEvent.PROCESS_BOOTED)
        logger.info(log_prefix + "connecting ...")
        self._on_lifecycle_event(ServerLifecycleEvent.CONNECTING)
        self._osc_protocol.connect(
            ip_address=self._options.ip_address,
            port=self._options.port,
            healthcheck=DEFAULT_HEALTHCHECK,
        )
        try:
            self._setup_notifications()
            self._contexts.add(self)
            self._osc_protocol.activate_healthcheck()
            self._setup_allocators()
            if self._client_id == 0:
                self._setup_system()
                self.sync()
            if self._osc_protocol.boot_future.result():
                self._boot_status = BootStatus.ONLINE
                self._on_lifecycle_event(ServerLifecycleEvent.OSC_CONNECTED)
                self._on_lifecycle_event(ServerLifecycleEvent.CONNECTED)
                logger.info(log_prefix + "... connected!")
                if owned:
                    logger.info(log_prefix + "... booted!")
                    self._on_lifecycle_event(ServerLifecycleEvent.BOOTED)
                self._boot_future.set_result(True)
        except TooManyClients:
            print("B")
            self._shutdown_future.set_result(ServerShutdownEvent.TOO_MANY_CLIENTS)
        # await shutdown future, osc panic, process panic
        shutdown = self._shutdown_future.result()
        # fire off quitting?
        if shutdown == ServerShutdownEvent.QUIT:
            logger.info(log_prefix + "quitting ...")
        self._boot_status = BootStatus.QUITTING
        if shutdown == ServerShutdownEvent.QUIT:
            try:
                Quit().communicate(server=self, timeout=1)
            except (OscProtocolOffline, asyncio.TimeoutError):
                pass
        elif shutdown == ServerShutdownEvent.DISCONNECT:
            pass
        elif shutdown == ServerShutdownEvent.OSC_PANIC:
            self._on_lifecycle_event(ServerLifecycleEvent.OSC_PANICKED)
        elif shutdown == ServerShutdownEvent.PROCESS_PANIC:
            self._on_lifecycle_event(ServerLifecycleEvent.PROCESS_PANICKED)
            self._on_lifecycle_event(ServerLifecycleEvent.OSC_PANICKED)
        if owned or shutdown == ServerShutdownEvent.QUIT:
            self._on_lifecycle_event(ServerLifecycleEvent.QUITTING)
        # handle shutdown future specifically with Quit request
        logger.info(log_prefix + "disconnecting ...")
        self._on_lifecycle_event(ServerLifecycleEvent.DISCONNECTING)
        self._osc_protocol.disconnect()
        if shutdown in (ServerShutdownEvent.QUIT, ServerShutdownEvent.DISCONNECT):
            self._on_lifecycle_event(ServerLifecycleEvent.OSC_DISCONNECTED)
        if owned:
            self._teardown_shm()
        self._teardown_state()
        self._is_owner = False
        if self in self._contexts:
            self._contexts.remove(self)
        logger.info(log_prefix + "disconnected!")
        self._on_lifecycle_event(ServerLifecycleEvent.DISCONNECTED)
        if owned:
            self._process_protocol.quit()
            if shutdown == ServerShutdownEvent.QUIT:
                self._on_lifecycle_event(ServerLifecycleEvent.PROCESS_QUIT)
        self._boot_status = BootStatus.OFFLINE
        if not self._boot_future.done():
            self._boot_future.set_result(
                shutdown in (ServerShutdownEvent.QUIT, ServerShutdownEvent.DISCONNECT)
            )
        if owned or shutdown == ServerShutdownEvent.QUIT:
            logger.info(log_prefix + "quit!")
            self._on_lifecycle_event(ServerLifecycleEvent.QUIT)
        self._exit_future.set_result(
            shutdown in (ServerShutdownEvent.QUIT, ServerShutdownEvent.DISCONNECT)
        )
        return

    def _on_lifecycle_event(self, event: ServerLifecycleEvent) -> None:
        for callback in self._lifecycle_event_callbacks.get(event, []):
            callback(event)

    def _setup_notifications(self) -> None:
        logger.info(self._log_prefix() + "setting up notifications ...")
        response = ToggleNotifications(True).communicate(server=self)
        if response is None or not isinstance(response, (DoneInfo, FailInfo)):
            raise RuntimeError
        if isinstance(response, FailInfo):
            raise TooManyClients
        if len(response.other) == 1:  # supernova doesn't provide a max logins value
            self._client_id = int(response.other[0])
            self._maximum_logins = self._options.maximum_logins
        else:
            self._client_id = int(response.other[0])
            self._maximum_logins = int(response.other[1])

    def _shutdown(self):
        if self.is_owner:
            self.quit()
        else:
            self.disconnect()

    ### PUBLIC METHODS ###

    def boot(self, *, options: Optional[Options] = None, **kwargs) -> "Server":
        """
        Boot the server.

        :param options: The context's options.
        :param kwargs: Keyword arguments for options.
        """
        if self._boot_status != BootStatus.OFFLINE:
            raise ServerOnline
        self._boot_status = BootStatus.BOOTING
        self._options = new(options or self._options, **kwargs)
        self._boot_future = concurrent.futures.Future()
        self._exit_future = concurrent.futures.Future()
        self._shutdown_future = concurrent.futures.Future()
        self._lifecycle_thread = threading.Thread(
            daemon=True,
            kwargs=dict(owned=True),
            target=self._lifecycle,
        )
        self._lifecycle_thread.start()
        if not (self._boot_future.result()):
            if (self._shutdown_future.result()) == ServerShutdownEvent.PROCESS_PANIC:
                raise ServerCannotBoot
        return self

    def connect(self, *, options: Optional[Options] = None, **kwargs) -> "Server":
        """
        Connect to a running server.

        :param options: The context's options.
        :param kwargs: Keyword arguments for options.
        """
        if self._boot_status != BootStatus.OFFLINE:
            raise ServerOnline
        self._boot_status = BootStatus.BOOTING
        self._options = new(options or self._options, **kwargs)
        self._boot_future = concurrent.futures.Future()
        self._exit_future = concurrent.futures.Future()
        self._shutdown_future = concurrent.futures.Future()
        self._lifecycle_thread = threading.Thread(
            daemon=True,
            kwargs=dict(owned=False),
            target=self._lifecycle,
        )
        self._lifecycle_thread.start()
        if not (self._boot_future.result()):
            if self._shutdown_future.result() == ServerShutdownEvent.TOO_MANY_CLIENTS:
                raise TooManyClients
        return self

    def disconnect(self) -> "Server":
        """
        Disconnect from a running server.
        """
        if self._boot_status == BootStatus.OFFLINE:
            raise ServerOffline
        if self._is_owner:
            raise OwnedServerShutdown("Cannot disconnect from owned server.")
        self._shutdown_future.set_result(ServerShutdownEvent.DISCONNECT)
        self._exit_future.result()
        return self

    def dump_tree(
        self,
        group: Optional[Group] = None,
        include_controls: bool = True,
        sync: bool = True,
    ) -> Optional[QueryTreeGroup]:
        """
        Dump the server's node tree.

        Emit ``/g_dumpTree`` requests.

        :param group: The group whose tree to query. Defaults to the root node.
        :param include_controls: Flag for including synth control values.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        self._validate_can_request()
        request = DumpTree(items=[(group or self.root_node, bool(include_controls))])
        if sync:
            with self.process_protocol.capture() as transcript:
                request.communicate(server=self)
                self.sync(timeout=10.0)
                return QueryTreeGroup.from_string("\n".join(transcript.lines))
        self._add_requests(request)
        return None

    def get_buffer(
        self, buffer: Buffer, *indices: int, sync: bool = True
    ) -> Optional[Dict[int, float]]:
        """
        Get a buffer sample.

        Emit ``/b_get`` requests.

        :param buffer: The buffer whose sample to get.
        :param indices: The sample indices to read.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = GetBuffer(buffer_id=buffer, indices=indices)
        if sync:
            return dict(cast(GetBufferInfo, request.communicate(server=self)).items)
        self._add_requests(request)
        return None

    def get_buffer_range(
        self, buffer: Buffer, index: int, count: int, sync: bool = True
    ) -> Optional[Sequence[float]]:
        """
        Get a buffer sample range.

        Emit ``/b_getn`` requests.

        :param buffer: The buffer whose samples to get.
        :param index: The sample index to start reading at.
        :param count: The number of samples to read.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = GetBufferRange(buffer_id=buffer, items=[(index, count)])
        if sync:
            return cast(GetBufferRangeInfo, request.communicate(server=self)).items[0][
                -1
            ]
        self._add_requests(request)
        return None

    def get_bus(self, bus: Bus, sync: bool = True) -> Optional[float]:
        """
        Get a control bus value.

        Emit ``/c_get`` requests.

        :param bus: The control bus whose value to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        if bus.calculation_rate != CalculationRate.CONTROL:
            raise InvalidCalculationRate
        request = GetControlBus(bus_ids=[bus.id_])
        if sync:
            return cast(GetControlBusInfo, request.communicate(server=self)).items[0][
                -1
            ]
        self._add_requests(request)
        return None

    def get_bus_range(
        self, bus: Bus, count: int, sync: bool = True
    ) -> Optional[Sequence[float]]:
        """
        Get a range of control bus values.

        Emit ``/c_getn`` requests.

        :param bus: The control bus to start reading at.
        :param count: The number of contiguous buses whose values to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        if bus.calculation_rate != CalculationRate.CONTROL:
            raise InvalidCalculationRate
        request = GetControlBusRange(items=[(bus.id_, count)])
        if sync:
            return cast(GetControlBusRangeInfo, request.communicate(server=self)).items[
                0
            ][-1]
        self._add_requests(request)
        return None

    def get_synth_controls(
        self, synth: Synth, *controls: Union[int, str], sync: bool = True
    ) -> Optional[Dict[Union[int, str], float]]:
        """
        Get a synth control.

        Emit ``/s_get`` requests.

        :param synth: The synth whose control to get.
        :param controls: The controls to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = GetSynthControl(synth_id=synth, controls=controls)
        if sync:
            return dict(
                cast(GetNodeControlInfo, request.communicate(server=self)).items
            )
        self._add_requests(request)
        return None

    def get_synth_control_range(
        self, synth: Synth, control: Union[int, str], count: int, sync: bool = True
    ) -> Optional[Sequence[Union[float, str]]]:
        """
        Get a range of synth controls.

        Emit ``/s_getn`` requests.

        :param synth: The synth whose control to get.
        :param control: The control to start reading at.
        :param count: The number of contiguous controls to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = GetSynthControlRange(synth_id=synth, items=[(control, count)])
        if sync:
            return cast(
                GetNodeControlRangeInfo, request.communicate(server=self)
            ).items[0][-1]
        self._add_requests(request)
        return None

    def query_buffer(self, buffer: Buffer, sync: bool = True) -> Optional[BufferInfo]:
        """
        Query a buffer.

        Emit ``/b_query`` requests.

        :param buffer: The buffer to query.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryBuffer(buffer_ids=[buffer.id_])
        if sync:
            return cast(BufferInfo, request.communicate(server=self))
        self._add_requests(request)
        return None

    def query_node(self, node: Node, sync: bool = True) -> Optional[NodeInfo]:
        """
        Query a node.

        Emit ``/n_query`` requests.

        :param node: The node to query.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryNode(node_ids=[node.id_])
        if sync:
            return cast(NodeInfo, request.communicate(server=self))
        self._add_requests(request)
        return None

    def query_status(self, sync: bool = True) -> Optional[StatusInfo]:
        """
        Query the server's status.

        Emit ``/status`` requests.

        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryStatus()
        if sync:
            return cast(StatusInfo, request.communicate(server=self))
        self._add_requests(request)
        return None

    def query_tree(
        self,
        group: Optional[Group] = None,
        include_controls: bool = True,
        sync: bool = True,
    ) -> Optional[QueryTreeGroup]:
        """
        Query the server's node tree.

        Emit ``/g_queryTree`` requests.

        :param group: The group whose tree to query. Defaults to the root node.
        :param include_controls: Flag for including synth control values.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryTree(items=[(group or 0, include_controls)])
        if sync:
            return QueryTreeGroup.from_query_tree_info(
                cast(QueryTreeInfo, request.communicate(server=self))
            )
        self._add_requests(request)
        return None

    def query_version(self, sync: bool = True) -> Optional[VersionInfo]:
        """
        Query the server's version.

        Emit ``/version`` requests.

        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryVersion()
        if sync:
            return cast(VersionInfo, request.communicate(server=self))
        self._add_requests(request)
        return None

    def quit(self, force: bool = False) -> "Server":
        """
        Quit the server.

        Emit ``/quit`` requests.

        :param force: Force the server to quit, even if this client doesn't own the
            process.
        """
        if self._boot_status != BootStatus.ONLINE:
            return self
        if not self._is_owner and not force:
            raise UnownedServerShutdown(
                "Cannot quit unowned server without force flag."
            )
        self._shutdown_future.set_result(ServerShutdownEvent.QUIT)
        self._exit_future.result()
        return self

    def reboot(self) -> "Server":
        """
        Reboot the server.
        """
        self.quit()
        self.boot()
        return self

    def reset(self) -> "Server":
        """
        Reset the server's state without quitting.
        """
        with self.at():
            self.clear_schedule()
            self.free_group_children(self.root_node)
            self.free_all_synthdefs()
        self.sync()
        self._teardown_state()
        self._setup_allocators()
        self._setup_system()
        self.sync()
        return self

    def sync(self, sync_id: Optional[int] = None, timeout: float = 1.0) -> "Server":
        """
        Sync the server.

        Emit ``/sync`` requests.

        :param sync_id: The sync ID to wait on.
        """
        if self._boot_status not in (BootStatus.BOOTING, BootStatus.ONLINE):
            raise ServerOffline
        Sync(
            sync_id=sync_id if sync_id is not None else self._get_next_sync_id()
        ).communicate(server=self, timeout=timeout)
        return self

    ### PUBLIC PROPERTIES ###

    @property
    def boot_future(self) -> concurrent.futures.Future[bool]:
        """
        Get the server's boot future.

        Only reference this _after_ booting or connecting, as the future is
        created when booting or connecting.
        """
        return self._boot_future

    @property
    def exit_future(self) -> concurrent.futures.Future[bool]:
        """
        Get the server's exit future.

        Only reference this _after_ booting or connecting, as the future is
        created when booting or connecting.
        """
        return self._exit_future

    @property
    def osc_protocol(self) -> ThreadedOscProtocol:
        """
        Get the server's OSC protocol.
        """
        return self._osc_protocol

    @property
    def process_protocol(self) -> SyncProcessProtocol:
        """
        Get the server's process protocol.
        """
        return self._process_protocol


class AsyncServer(BaseServer):
    """
    A realtime execution context with :py:mod:`asyncio`-based OSC and process protocols.

    :param options: The context's options.
    :param kwargs: Keyword arguments for options.
    """

    ### INITIALIZER ###

    def __init__(
        self, options: Optional[Options] = None, name: Optional[str] = None, **kwargs
    ):
        super().__init__(
            name=name,
            options=options,
            **kwargs,
        )
        self._boot_future: asyncio.Future[bool] = asyncio.Future()
        self._exit_future: asyncio.Future[bool] = asyncio.Future()
        self._shutdown_future: asyncio.Future[ServerShutdownEvent] = asyncio.Future()
        self._osc_protocol = AsyncOscProtocol(
            name=name,
            on_panic_callback=lambda: self._shutdown_future.set_result(
                ServerShutdownEvent.OSC_PANIC
            ),
        )
        self._process_protocol = AsyncProcessProtocol(
            name=name,
            on_panic_callback=lambda: self._shutdown_future.set_result(
                ServerShutdownEvent.PROCESS_PANIC
            ),
        )
        self._setup_osc_callbacks(self._osc_protocol)

    ### PRIVATE METHODS ###

    async def _lifecycle(self, owned=True) -> None:
        log_prefix = self._log_prefix()
        logger.info(log_prefix + "booting ...")
        if owned:
            await self._on_lifecycle_event(ServerLifecycleEvent.BOOTING)
            try:
                await self._process_protocol.boot(self._options)
            except ServerCannotBoot:
                await self._on_lifecycle_event(ServerLifecycleEvent.PROCESS_PANICKED)
                self._boot_status = BootStatus.OFFLINE
                self._boot_future.set_result(False)
                self._exit_future.set_result(False)
                return
            self._is_owner = True
            self._setup_shm()
            await self._on_lifecycle_event(ServerLifecycleEvent.PROCESS_BOOTED)
        logger.info(log_prefix + "connecting ...")
        await self._on_lifecycle_event(ServerLifecycleEvent.CONNECTING)
        await self._osc_protocol.connect(
            ip_address=self._options.ip_address,
            port=self._options.port,
            healthcheck=DEFAULT_HEALTHCHECK,
        )
        try:
            await self._setup_notifications()
            self._contexts.add(self)
            self._osc_protocol.activate_healthcheck()
            self._setup_allocators()
            if self._client_id == 0:
                self._setup_system()
                await self.sync()
            if await self._osc_protocol.boot_future:
                self._boot_status = BootStatus.ONLINE
                await self._on_lifecycle_event(ServerLifecycleEvent.OSC_CONNECTED)
                await self._on_lifecycle_event(ServerLifecycleEvent.CONNECTED)
                logger.info(log_prefix + "... connected!")
                if owned:
                    logger.info(log_prefix + "... booted!")
                    await self._on_lifecycle_event(ServerLifecycleEvent.BOOTED)
                self._boot_future.set_result(True)
        except TooManyClients:
            self._shutdown_future.set_result(ServerShutdownEvent.TOO_MANY_CLIENTS)
        # await shutdown future, osc panic, process panic
        shutdown = await self._shutdown_future
        # fire off quitting?
        if shutdown == ServerShutdownEvent.QUIT:
            logger.info(log_prefix + "quitting ...")
        self._boot_status = BootStatus.QUITTING
        if shutdown == ServerShutdownEvent.QUIT:
            try:
                await Quit().communicate_async(server=self, timeout=1)
            except (OscProtocolOffline, asyncio.TimeoutError):
                pass
        elif shutdown == ServerShutdownEvent.DISCONNECT:
            pass
        elif shutdown == ServerShutdownEvent.OSC_PANIC:
            await self._on_lifecycle_event(ServerLifecycleEvent.OSC_PANICKED)
        elif shutdown == ServerShutdownEvent.PROCESS_PANIC:
            await self._on_lifecycle_event(ServerLifecycleEvent.PROCESS_PANICKED)
            await self._on_lifecycle_event(ServerLifecycleEvent.OSC_PANICKED)
        if owned or shutdown == ServerShutdownEvent.QUIT:
            await self._on_lifecycle_event(ServerLifecycleEvent.QUITTING)
        # handle shutdown future specifically with Quit request
        logger.info(log_prefix + "disconnecting ...")
        await self._on_lifecycle_event(ServerLifecycleEvent.DISCONNECTING)
        await self._osc_protocol.disconnect()
        if shutdown in (ServerShutdownEvent.QUIT, ServerShutdownEvent.DISCONNECT):
            await self._on_lifecycle_event(ServerLifecycleEvent.OSC_DISCONNECTED)
        if owned:
            self._teardown_shm()
        self._teardown_state()
        self._is_owner = False
        if self in self._contexts:
            self._contexts.remove(self)
        logger.info(log_prefix + "disconnected!")
        await self._on_lifecycle_event(ServerLifecycleEvent.DISCONNECTED)
        if owned:
            await self._process_protocol.quit()
            if shutdown == ServerShutdownEvent.QUIT:
                await self._on_lifecycle_event(ServerLifecycleEvent.PROCESS_QUIT)
        self._boot_status = BootStatus.OFFLINE
        if not self._boot_future.done():
            self._boot_future.set_result(
                shutdown in (ServerShutdownEvent.QUIT, ServerShutdownEvent.DISCONNECT)
            )
        if owned or shutdown == ServerShutdownEvent.QUIT:
            logger.info(log_prefix + "quit!")
            await self._on_lifecycle_event(ServerLifecycleEvent.QUIT)
        self._exit_future.set_result(
            shutdown in (ServerShutdownEvent.QUIT, ServerShutdownEvent.DISCONNECT)
        )
        return

    async def _on_lifecycle_event(self, event: ServerLifecycleEvent) -> None:
        for callback in self._lifecycle_event_callbacks.get(event, []):
            if asyncio.iscoroutine(result := callback(event)):
                await result

    async def _setup_notifications(self) -> None:
        logger.info(self._log_prefix() + "setting up notifications ...")
        response = await ToggleNotifications(True).communicate_async(server=self)
        if response is None or not isinstance(response, (DoneInfo, FailInfo)):
            raise RuntimeError
        if isinstance(response, FailInfo):
            raise TooManyClients
        if len(response.other) == 1:  # supernova doesn't provide a max logins value
            self._client_id = int(response.other[0])
            self._maximum_logins = self._options.maximum_logins
        else:
            self._client_id = int(response.other[0])
            self._maximum_logins = int(response.other[1])

    async def _shutdown(self):
        if self.is_owner:
            await self.quit()
        else:
            await self.disconnect()

    ### PUBLIC METHODS ###

    async def boot(
        self, *, options: Optional[Options] = None, **kwargs
    ) -> "AsyncServer":
        """
        Boot the server.

        :param options: The context's options.
        :param kwargs: Keyword arguments for options.
        """
        if self._boot_status != BootStatus.OFFLINE:
            raise ServerOnline
        self._boot_status = BootStatus.BOOTING
        self._options = new(options or self._options, **kwargs)
        loop = asyncio.get_running_loop()
        self._boot_future = loop.create_future()
        self._exit_future = loop.create_future()
        self._shutdown_future = loop.create_future()
        self._lifecycle_task = loop.create_task(self._lifecycle(owned=True))
        if not (await self._boot_future):
            if (await self._shutdown_future) == ServerShutdownEvent.PROCESS_PANIC:
                raise ServerCannotBoot
        return self

    async def connect(
        self, *, options: Optional[Options] = None, **kwargs
    ) -> "AsyncServer":
        """
        Connect to a running server.

        :param options: The context's options.
        :param kwargs: Keyword arguments for options.
        """
        if self._boot_status != BootStatus.OFFLINE:
            raise ServerOnline
        self._boot_status = BootStatus.BOOTING
        self._options = new(options or self._options, **kwargs)
        loop = asyncio.get_running_loop()
        self._boot_future = loop.create_future()
        self._exit_future = loop.create_future()
        self._shutdown_future = loop.create_future()
        self._lifecycle_task = loop.create_task(self._lifecycle(owned=False))
        if not (await self._boot_future):
            if await self._shutdown_future == ServerShutdownEvent.TOO_MANY_CLIENTS:
                raise TooManyClients
        return self

    async def disconnect(self) -> "AsyncServer":
        """
        Disconnect from a running server.
        """
        if self._boot_status == BootStatus.OFFLINE:
            raise ServerOffline
        if self._is_owner:
            raise OwnedServerShutdown("Cannot disconnect from owned server.")
        self._shutdown_future.set_result(ServerShutdownEvent.DISCONNECT)
        await self._exit_future
        return self

    async def dump_tree(
        self,
        group: Optional[Group] = None,
        include_controls: bool = True,
        sync: bool = True,
    ) -> Optional[QueryTreeGroup]:
        """
        Dump the server's node tree.

        Emit ``/g_dumpTree`` requests.

        :param group: The group whose tree to query. Defaults to the root node.
        :param include_controls: Flag for including synth control values.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        self._validate_can_request()
        request = DumpTree(items=[(group or self.root_node, bool(include_controls))])
        if sync:
            with self.process_protocol.capture() as transcript:
                await request.communicate_async(server=self)
                await self.sync(timeout=10.0)
                return QueryTreeGroup.from_string("\n".join(transcript.lines))
        self._add_requests(request)
        return None

    async def get_buffer(
        self, buffer: Buffer, *indices: int, sync: bool = True
    ) -> Optional[Dict[int, float]]:
        """
        Get a buffer sample.

        Emit ``/b_get`` requests.

        :param buffer: The buffer whose sample to get.
        :param indices: The sample indices to read.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = GetBuffer(buffer_id=buffer, indices=indices)
        if sync:
            return dict(
                cast(GetBufferInfo, await request.communicate_async(server=self)).items
            )
        self._add_requests(request)
        return None

    async def get_buffer_range(
        self, buffer: Buffer, index: int, count: int, sync: bool = True
    ) -> Optional[Sequence[float]]:
        """
        Get a buffer sample range.

        Emit ``/b_getn`` requests.

        :param buffer: The buffer whose samples to get.
        :param index: The sample index to start reading at.
        :param count: The number of samples to read.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = GetBufferRange(buffer_id=buffer, items=[(index, count)])
        if sync:
            return cast(
                GetBufferRangeInfo, await request.communicate_async(server=self)
            ).items[0][-1]
        self._add_requests(request)
        return None

    async def get_bus(self, bus: Bus, sync: bool = True) -> Optional[float]:
        """
        Get a control bus value.

        Emit ``/c_get`` requests.

        :param bus: The control bus whose value to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        if bus.calculation_rate != CalculationRate.CONTROL:
            raise InvalidCalculationRate
        request = GetControlBus(bus_ids=[bus.id_])
        if sync:
            return cast(
                GetControlBusInfo, await request.communicate_async(server=self)
            ).items[0][-1]
        self._add_requests(request)
        return None

    async def get_bus_range(
        self, bus: Bus, count: int, sync: bool = True
    ) -> Optional[Sequence[float]]:
        """
        Get a range of control bus values.

        Emit ``/c_getn`` requests.

        :param bus: The control bus to start reading at.
        :param count: The number of contiguous buses whose values to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        if bus.calculation_rate != CalculationRate.CONTROL:
            raise InvalidCalculationRate
        request = GetControlBusRange(items=[(bus.id_, count)])
        if sync:
            return cast(
                GetControlBusRangeInfo, await request.communicate_async(server=self)
            ).items[0][-1]
        self._add_requests(request)
        return None

    async def get_synth_controls(
        self, synth: Synth, *controls: Union[int, str], sync: bool = True
    ) -> Optional[Dict[Union[int, str], float]]:
        """
        Get a synth control.

        Emit ``/s_get`` requests.

        :param synth: The synth whose control to get.
        :param controls: The controls to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = GetSynthControl(synth_id=synth, controls=controls)
        if sync:
            return dict(
                cast(
                    GetNodeControlInfo, await request.communicate_async(server=self)
                ).items
            )
        self._add_requests(request)
        return None

    async def get_synth_control_range(
        self, synth: Synth, control: Union[int, str], count: int, sync: bool = True
    ) -> Optional[Sequence[Union[float, str]]]:
        """
        Get a range of synth controls.

        Emit ``/s_getn`` requests.

        :param synth: The synth whose control to get.
        :param control: The control to start reading at.
        :param count: The number of contiguous controls to get.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = GetSynthControlRange(synth_id=synth, items=[(control, count)])
        if sync:
            return cast(
                GetNodeControlRangeInfo, await request.communicate_async(server=self)
            ).items[0][-1]
        self._add_requests(request)
        return None

    async def query_buffer(
        self, buffer: Buffer, sync: bool = True
    ) -> Optional[BufferInfo]:
        """
        Query a buffer.

        Emit ``/b_query`` requests.

        :param buffer: The buffer to query.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryBuffer(buffer_ids=[buffer.id_])
        if sync:
            return cast(BufferInfo, await request.communicate_async(server=self))
        self._add_requests(request)
        return None

    async def query_node(self, node: Node, sync: bool = True) -> Optional[NodeInfo]:
        """
        Query a node.

        Emit ``/n_query`` requests.

        :param node: The node to query.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryNode(node_ids=[node.id_])
        if sync:
            return cast(NodeInfo, await request.communicate_async(server=self))
        self._add_requests(request)
        return None

    async def query_status(self, sync: bool = True) -> Optional[StatusInfo]:
        """
        Query the server's status.

        Emit ``/status`` requests.

        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryStatus()
        if sync:
            return cast(StatusInfo, await request.communicate_async(server=self))
        self._add_requests(request)
        return None

    async def query_tree(
        self,
        group: Optional[Group] = None,
        include_controls: bool = True,
        sync: bool = True,
    ) -> Optional[QueryTreeGroup]:
        """
        Query the server's node tree.

        Emit ``/g_queryTree`` requests.

        :param group: The group whose tree to query. Defaults to the root node.
        :param include_controls: Flag for including synth control values.
        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryTree(items=[(group or 0, include_controls)])
        if sync:
            return QueryTreeGroup.from_query_tree_info(
                cast(QueryTreeInfo, await request.communicate_async(server=self))
            )
        self._add_requests(request)
        return None

    async def query_version(self, sync: bool = True) -> Optional[VersionInfo]:
        """
        Query the server's version.

        Emit ``/version`` requests.

        :param sync: If true, communicate the request immediately. Otherwise bundle it
            with the current request context.
        """
        request = QueryVersion()
        if sync:
            return cast(VersionInfo, await request.communicate_async(server=self))
        self._add_requests(request)
        return None

    async def quit(self, force: bool = False) -> "AsyncServer":
        """
        Quit the server.

        Emit ``/quit`` requests.

        :param force: Force the server to quit, even if this client doesn't own the
            process.
        """
        if self._boot_status != BootStatus.ONLINE:
            return self
        if not self._is_owner and not force:
            raise UnownedServerShutdown(
                "Cannot quit unowned server without force flag."
            )
        self._shutdown_future.set_result(ServerShutdownEvent.QUIT)
        await self._exit_future
        return self

    async def reboot(self) -> "AsyncServer":
        """
        Reboot the server.
        """
        await self.quit()
        await self.boot()
        return self

    async def reset(self) -> "AsyncServer":
        """
        Reset the server's state without quitting.
        """
        with self.at():
            self.clear_schedule()
            self.free_all_synthdefs()
            self.free_group_children(self.root_node)
        await self.sync()
        self._teardown_state()
        self._setup_allocators()
        self._setup_system()
        await self.sync()
        return self

    async def sync(
        self, sync_id: Optional[int] = None, timeout: float = 1.0
    ) -> "AsyncServer":
        """
        Sync the server.

        Emit ``/sync`` requests.

        :param sync_id: The sync ID to wait on.
        """
        if self._boot_status not in (BootStatus.BOOTING, BootStatus.ONLINE):
            raise ServerOffline
        await Sync(
            sync_id=sync_id if sync_id is not None else self._get_next_sync_id()
        ).communicate_async(server=self, timeout=timeout)
        return self

    ### PUBLIC PROPERTIES ###

    @property
    def boot_future(self) -> asyncio.Future[bool]:
        """
        Get the server's boot future.

        Only reference this _after_ booting or connecting, as the future is
        created when booting or connecting.
        """
        return self._boot_future

    @property
    def exit_future(self) -> asyncio.Future[bool]:
        """
        Get the server's exit future.

        Only reference this _after_ booting or connecting, as the future is
        created when booting or connecting.
        """
        return self._exit_future

    @property
    def osc_protocol(self) -> AsyncOscProtocol:
        """
        Get the server's OSC protocol.
        """
        return self._osc_protocol

    @property
    def process_protocol(self) -> AsyncProcessProtocol:
        """
        Get the server's process protocol.
        """
        return self._process_protocol



from typing import Any, Dict, Tuple

from ..enums import CalculationRate, SignalRange
from .core import UGen, param, ugen


@ugen(ar=True, kr=True)
class BrownNoise(UGen):
    """
    A brown noise unit generator.

    ::

        >>> supriya.ugens.BrownNoise.ar()
        <BrownNoise.ar()[0]>
    """


@ugen(ar=True, kr=True)
class ClipNoise(UGen):
    """
    A clipped noise unit generator.

    ::

        >>> supriya.ugens.ClipNoise.ar()
        <ClipNoise.ar()[0]>
    """


@ugen(ar=True, kr=True)
class CoinGate(UGen):
    """
    A probabilistic trigger gate.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> coin_gate = supriya.ugens.CoinGate.ar(
        ...     probability=0.5,
        ...     trigger=trigger,
        ... )
        >>> coin_gate
        <CoinGate.ar()[0]>
    """

    probability = param(0.5)
    trigger = param()


@ugen(ar=True, kr=True)
class Crackle(UGen):
    """
    A chaotic noise generator.

    ::

        >>> crackle = supriya.ugens.Crackle.ar(
        ...     chaos_parameter=1.25,
        ... )
        >>> crackle
        <Crackle.ar()[0]>
    """

    chaos_parameter = param(1.5)


@ugen(ar=True, kr=True, signal_range=SignalRange.UNIPOLAR)
class Dust(UGen):
    """
    A unipolar random impulse generator.

    ::

        >>> dust = supriya.ugens.Dust.ar(
        ...     density=23,
        ... )
        >>> dust
        <Dust.ar()[0]>
    """

    density = param(0.0)


@ugen(ar=True, kr=True)
class Dust2(UGen):
    """
    A bipolar random impulse generator.

    ::

        >>> dust_2 = supriya.ugens.Dust2.ar(
        ...     density=23,
        ... )
        >>> dust_2
        <Dust2.ar()[0]>
    """

    density = param(0.0)


@ugen(ir=True)
class ExpRand(UGen):
    """
    An exponential random distribution.

    ::

        >>> exp_rand = supriya.ugens.ExpRand.ir()
        >>> exp_rand
        <ExpRand.ir()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs,
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        if isinstance(kwargs["minimum"], float) and isinstance(
            kwargs["maximum"], float
        ):
            kwargs["minimum"], kwargs["maximum"] = sorted(
                [kwargs["minimum"], kwargs["maximum"]]
            )
        return calculation_rate, kwargs


@ugen(ar=True, kr=True)
class GrayNoise(UGen):
    """
    A gray noise unit generator.

    ::

        >>> supriya.ugens.GrayNoise.ar()
        <GrayNoise.ar()[0]>
    """


@ugen(ar=True, kr=True)
class Hasher(UGen):
    """
    A signal hasher.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> hasher = supriya.ugens.Hasher.ar(
        ...     source=source,
        ... )
        >>> hasher
        <Hasher.ar()[0]>
    """

    source = param()


@ugen(ir=True)
class IRand(UGen):
    """
    An integer uniform random distribution.

    ::

        >>> supriya.ugens.IRand.ir()
        <IRand.ir()[0]>
    """

    minimum = param(0)
    maximum = param(127)


@ugen(ar=True, kr=True)
class LFClipNoise(UGen):
    """
    A dynamic clipped noise generator.

    ::

        >>> supriya.ugens.LFClipNoise.ar()
        <LFClipNoise.ar()[0]>
    """

    frequency = param(500.0)


@ugen(ar=True, kr=True)
class LFDClipNoise(UGen):
    """
    A clipped noise generator.

    ::

        >>> supriya.ugens.LFDClipNoise.ar()
        <LFDClipNoise.ar()[0]>
    """

    frequency = param(500.0)


@ugen(ar=True, kr=True)
class LFDNoise0(UGen):
    """
    A dynamic step noise generator.

    ::

        >>> supriya.ugens.LFDNoise0.ar()
        <LFDNoise0.ar()[0]>
    """

    frequency = param(500.0)


@ugen(ar=True, kr=True)
class LFDNoise1(UGen):
    """
    A dynamic ramp noise generator.

    ::

        >>> supriya.ugens.LFDNoise1.ar()
        <LFDNoise1.ar()[0]>
    """

    frequency = param(500.0)


@ugen(ar=True, kr=True)
class LFDNoise3(UGen):
    """
    A dynamic polynomial noise generator.

    ::

        >>> supriya.ugens.LFDNoise3.ar()
        <LFDNoise3.ar()[0]>
    """

    frequency = param(500.0)


@ugen(ar=True, kr=True)
class LFNoise0(UGen):
    """
    A step noise generator.

    ::

        >>> supriya.ugens.LFNoise0.ar()
        <LFNoise0.ar()[0]>
    """

    frequency = param(500.0)


@ugen(ar=True, kr=True)
class LFNoise1(UGen):
    """
    A ramp noise generator.

    ::

        >>> supriya.ugens.LFNoise1.ar()
        <LFNoise1.ar()[0]>
    """

    frequency = param(500.0)


@ugen(ar=True, kr=True)
class LFNoise2(UGen):
    """
    A quadratic noise generator.

    ::

        >>> supriya.ugens.LFNoise2.ar()
        <LFNoise2.ar()[0]>
    """

    frequency = param(500.0)


@ugen(ir=True)
class LinRand(UGen):
    """
    A skewed linear random distribution.

    ::

        >>> lin_rand = supriya.ugens.LinRand.ir(
        ...     minimum=-1.0,
        ...     maximum=1.0,
        ...     skew=0.5,
        ... )
        >>> lin_rand
        <LinRand.ir()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)
    skew = param(0)


@ugen(ar=True, kr=True)
class Logistic(UGen):
    """
    A chaotic noise function.

    ::

        >>> logistic = supriya.ugens.Logistic.ar(
        ...     chaos_parameter=3.0,
        ...     frequency=1000,
        ...     initial_y=0.5,
        ... )
        >>> logistic
        <Logistic.ar()[0]>
    """

    chaos_parameter = param(3)
    frequency = param(1000)
    initial_y = param(0.5)


@ugen(ar=True, kr=True)
class MantissaMask(UGen):
    """
    A floating-point mantissa mask.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> mantissa_mask = supriya.ugens.MantissaMask.ar(
        ...     source=source,
        ...     bits=3,
        ... )
        >>> mantissa_mask
        <MantissaMask.ar()[0]>
    """

    source = param(0)
    bits = param(3)


@ugen(ir=True)
class NRand(UGen):
    """
    A sum of `n` uniform distributions.

    ::

        >>> n_rand = supriya.ugens.NRand.ir(
        ...     minimum=-1,
        ...     maximum=1,
        ...     n=1,
        ... )
        >>> n_rand
        <NRand.ir()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)
    n = param(1)


@ugen(ar=True, kr=True)
class PinkNoise(UGen):
    """
    A pink noise unit generator.

    ::

        >>> supriya.ugens.PinkNoise.ar()
        <PinkNoise.ar()[0]>
    """


@ugen(ir=True)
class Rand(UGen):
    """
    A uniform random distribution.

    ::

        >>> supriya.ugens.Rand.ir()
        <Rand.ir()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)


@ugen(kr=True, ir=True, is_width_first=True)
class RandID(UGen):
    """
    Sets the synth's random generator ID.

    ::

        >>> rand_id = supriya.ugens.RandID.ir(
        ...     rand_id=1,
        ... )
        >>> rand_id
        <RandID.ir()[0]>
    """

    rand_id = param(1)


@ugen(ar=True, kr=True, ir=True, is_width_first=True)
class RandSeed(UGen):
    """
    Sets the synth's random generator seed.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> rand_seed = supriya.ugens.RandSeed.ar(
        ...     seed=1,
        ...     trigger=trigger,
        ... )
        >>> rand_seed
        <RandSeed.ar()[0]>
    """

    trigger = param(0)
    seed = param(56789)


@ugen(ar=True, kr=True)
class TExpRand(UGen):
    """
    A triggered exponential random number generator.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> t_exp_rand = supriya.ugens.TExpRand.ar(
        ...     minimum=-1.0,
        ...     maximum=1.0,
        ...     trigger=trigger,
        ... )
        >>> t_exp_rand
        <TExpRand.ar()[0]>
    """

    minimum = param(0.01)
    maximum = param(1.0)
    trigger = param(0)


@ugen(ar=True, kr=True)
class TIRand(UGen):
    """
    A triggered integer random number generator.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> t_i_rand = supriya.ugens.TIRand.ar(
        ...     minimum=0,
        ...     maximum=127,
        ...     trigger=trigger,
        ... )
        >>> t_i_rand
        <TIRand.ar()[0]>
    """

    minimum = param(0)
    maximum = param(127)
    trigger = param(0)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs,
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        kwargs["minimum"] = int(kwargs["minimum"])
        kwargs["maximum"] = int(kwargs["maximum"])
        return calculation_rate, kwargs


@ugen(ar=True, kr=True)
class TRand(UGen):
    """
    A triggered random number generator.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> t_rand = supriya.ugens.TRand.ar(
        ...     minimum=-1.0,
        ...     maximum=1.0,
        ...     trigger=trigger,
        ... )
        >>> t_rand
        <TRand.ar()[0]>
    """

    minimum = param(0.0)
    maximum = param(1.0)
    trigger = param(0)


@ugen(ar=True, kr=True)
class TWindex(UGen):
    """
    A triggered windex.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> t_windex = supriya.ugens.TWindex.ar(
        ...     trigger=trigger,
        ...     normalize=0,
        ...     array=[1, 2, 3],
        ... )
        >>> t_windex
        <TWindex.ar()[0]>
    """

    trigger = param()
    normalize = param(0)
    array = param(unexpanded=True)


@ugen(ar=True, kr=True)
class WhiteNoise(UGen):
    """
    A white noise unit generator.

    ::

        >>> supriya.ugens.WhiteNoise.ar()
        <WhiteNoise.ar()[0]>
    """



import collections
import dataclasses
import fractions
import itertools
import logging
import queue
import time
import traceback
from typing import Any, Callable, Deque, Dict, FrozenSet, Optional, Set, Tuple, Union

from .. import conversions
from .ephemera import (
    Action,
    CallbackCommand,
    CallbackEvent,
    ChangeCommand,
    ChangeEvent,
    ClockContext,
    ClockState,
    Command,
    Event,
    EventType,
    Moment,
    Quantization,
    TimeUnit,
)
from .eventqueue import EventQueue

logger = logging.getLogger(__name__)


class BaseClock:
    ### CLASS VARIABLES ###

    _valid_quantizations: FrozenSet[Quantization] = frozenset(
        [
            "8M",
            "4M",
            "2M",
            "1M",
            "1/2",
            "1/2T",
            "1/4",
            "1/4T",
            "1/8",
            "1/8T",
            "1/16",
            "1/16T",
            "1/32",
            "1/32T",
            "1/64",
            "1/64T",
            "1/128",
        ]
    )

    ### INITIALIZER ###

    def __init__(self) -> None:
        self._name = None
        self._counter = itertools.count()
        self._command_deque: Deque[Command] = collections.deque()
        self._event_queue = EventQueue()
        self._is_running = False
        self._slop = 0.001
        self._actions_by_id: Dict[int, Action] = {}
        self._measure_relative_event_ids: Set[int] = set()
        self._offset_relative_event_ids: Set[int] = set()
        self._state = ClockState(
            beats_per_minute=120.0,
            initial_seconds=0.0,
            previous_measure=1,
            previous_offset=0.0,
            previous_seconds=0.0,
            previous_time_signature_change_offset=0.0,
            time_signature=(4, 4),
        )

    ### TIME METHODS ###

    def _get_cue_point(
        self, seconds: float, quantization: Quantization
    ) -> Tuple[float, float, Optional[int]]:
        moment = self._seconds_to_moment(seconds)
        if quantization is None:
            offset: float = moment.offset
            measure: Optional[int] = None
        elif "M" in quantization:
            measure_grid = int(quantization[0])
            grid_offset = moment.measure - 1 + (moment.measure_offset > 0)
            count, modulus = divmod(grid_offset, measure_grid)
            count += modulus > 0
            measure = (count * measure_grid) + 1
            offset = self._measure_to_offset(measure)
        else:
            measure = None
            fraction_grid = self.quantization_to_beats(quantization)
            div, mod = divmod(moment.offset, fraction_grid)
            offset = float(div * fraction_grid)
            if mod:
                offset += fraction_grid
        seconds = self._offset_to_seconds(offset)
        logger.debug(
            f"[{self.name}] ... ... Cueing {quantization} to "
            f"{seconds}:s / {offset}:o / {measure}:m"
        )
        return seconds, offset, measure

    def _get_schedule_point(
        self, schedule_at: float, time_unit: TimeUnit
    ) -> Tuple[float, Optional[float], Optional[int]]:
        measure: Optional[int] = None
        offset: Optional[float] = None
        if time_unit == TimeUnit.MEASURES:
            measure = int(schedule_at)
            offset = self._measure_to_offset(measure)
            seconds: float = self._offset_to_seconds(offset)
        elif time_unit == TimeUnit.BEATS:
            offset = float(schedule_at)
            seconds = self._offset_to_seconds(schedule_at)
        else:
            seconds = float(schedule_at)
        return seconds, offset, measure

    def _measure_to_offset(self, measure: int) -> float:
        return conversions.measure_to_offset(
            measure,
            self._state.time_signature,
            self._state.previous_measure,
            self._state.previous_time_signature_change_offset,
        )

    def _offset_to_measure(self, offset: float) -> int:
        return conversions.offset_to_measure(
            offset,
            self._state.time_signature,
            self._state.previous_measure,
            self._state.previous_time_signature_change_offset,
        )

    def _offset_to_measure_offset(self, offset: float) -> float:
        return conversions.offset_to_measure_offset(
            offset,
            self._state.time_signature,
            self._state.previous_time_signature_change_offset,
        )

    def _offset_to_moment(self, offset: float) -> Moment:
        seconds = self._offset_to_seconds(offset)
        measure, measure_offset = divmod(
            offset - self._state.previous_time_signature_change_offset,
            self._state.time_signature[0] / self._state.time_signature[1],
        )
        return Moment(
            beats_per_minute=self._state.beats_per_minute,
            measure=int(measure + self._state.previous_measure),
            measure_offset=measure_offset,
            offset=offset,
            seconds=seconds,
            time_signature=self._state.time_signature,
        )

    def _offset_to_seconds(self, offset: float) -> float:
        return conversions.offset_to_seconds(
            beats_per_minute=self._state.beats_per_minute,
            current_offset=offset,
            previous_offset=self._state.previous_offset,
            previous_seconds=self._state.previous_seconds,
            beat_duration=1 / self._state.time_signature[1],
        )

    def _seconds_to_moment(self, seconds: float) -> Moment:
        offset = self._seconds_to_offset(seconds)
        measure, measure_offset = divmod(
            offset - self._state.previous_time_signature_change_offset,
            self._state.time_signature[0] / self._state.time_signature[1],
        )
        return Moment(
            beats_per_minute=self._state.beats_per_minute,
            measure=int(measure + self._state.previous_measure),
            measure_offset=measure_offset,
            offset=offset,
            seconds=seconds,
            time_signature=self._state.time_signature,
        )

    def _seconds_to_offset(self, seconds: float) -> float:
        return conversions.seconds_to_offset(
            beats_per_minute=self._state.beats_per_minute,
            current_time=seconds,
            previous_offset=self._state.previous_offset,
            previous_seconds=self._state.previous_seconds,
            beat_duration=1 / self._state.time_signature[1],
        )

    ### SCHEDULING METHODS ###

    def _cancel(self, event_id: int) -> Optional[Action]:
        action = self._actions_by_id.pop(event_id, None)
        if action is not None and isinstance(action, Event):
            self._event_queue.remove(action)
            if action.offset is not None:
                self._offset_relative_event_ids.remove(action.event_id)
                if action.measure is not None:
                    self._measure_relative_event_ids.remove(action.event_id)
        return action

    def _enqueue_command(self, command: Command) -> None:
        self._actions_by_id[command.event_id] = command
        self._command_deque.append(command)
        if isinstance(command, CallbackCommand):
            logger.debug(
                f"[{self.name}] Enqueued {type(command).__name__} ({command.event_id}) {command.procedure}"
            )
        else:
            logger.debug(
                f"[{self.name}] Enqueued {type(command).__name__} ({command.event_id})"
            )

    def _enqueue_event(self, event: Event) -> None:
        self._actions_by_id[event.event_id] = event
        self._event_queue.put(event)
        if event.offset is not None:
            self._offset_relative_event_ids.add(event.event_id)
            if event.measure is not None:
                self._measure_relative_event_ids.add(event.event_id)

    def _process_perform_event_loop(
        self, current_moment: Moment
    ) -> Tuple[Optional[Event], Optional[Moment], bool, bool]:
        # There may be items in the queue which have been flagged "removed".
        # They contribute to the queue's size, but cannot be retrieved by .get().
        try:
            event = self._event_queue.get()
        except queue.Empty:
            return None, None, True, False
        if self._actions_by_id.pop(event.event_id, None) is None:
            return None, None, True, False
        if current_moment.seconds < event.seconds:
            self._enqueue_event(event)
            return None, None, False, True
        if event.offset is not None:
            desired_moment = self._offset_to_moment(event.offset)
        else:
            desired_moment = self._seconds_to_moment(event.seconds)
        if event.offset is not None and event.offset != desired_moment.offset:
            raise RuntimeError(
                f"Offset mismatch: {event.offset} vs {desired_moment.offset}"
            )
        return event, desired_moment, False, False

    def _perform_callback_event(
        self, event: CallbackEvent, current_moment: Moment, desired_moment: Moment
    ) -> None:
        logger.debug(
            f"[{self.name}] ... ... Performing {event.procedure} at "
            f"{desired_moment.seconds - self._state.initial_seconds}:s / "
            f"{desired_moment.offset}:o"
        )
        context = ClockContext(current_moment, desired_moment, event)
        args = event.args or ()
        kwargs = event.kwargs or {}
        try:
            result = event.procedure(context, *args, **kwargs)
        except Exception:
            traceback.print_exc()
            return
        if isinstance(result, float) or result is None:
            delta, time_unit = result, TimeUnit.BEATS
        else:
            delta, time_unit = result
        self._process_callback_event_result(desired_moment, event, delta, time_unit)

    def _process_callback_event_result(
        self,
        desired_moment: Moment,
        event: CallbackEvent,
        delta: Optional[float],
        time_unit: TimeUnit,
    ) -> None:
        if delta is None or delta <= 0:
            return
        invocations = event.invocations + 1
        measure: Optional[int] = None
        offset: Optional[float] = None
        if time_unit in (TimeUnit.BEATS, TimeUnit.MEASURES):
            if time_unit == TimeUnit.MEASURES:
                measure = desired_moment.measure + int(delta)
                offset = self._measure_to_offset(measure)
            else:
                offset = desired_moment.offset + delta
            seconds = self._offset_to_seconds(offset)
        else:
            seconds = desired_moment.seconds + delta
        logger.debug(
            f"[{self.name}] ... ... ... Rescheduling "
            f"{event.procedure} ({event.event_id}) at {seconds - self._state.initial_seconds}s"
        )
        event = dataclasses.replace(
            event,
            invocations=invocations,
            measure=measure,
            offset=offset,
            seconds=seconds,
        )
        self._enqueue_event(event)

    def _perform_change_event(
        self, event: ChangeEvent, current_moment: Moment, desired_moment: Moment
    ) -> Tuple[Moment, bool]:
        logger.debug(
            f"[{self.name}] ... ... Changing at "
            f"{desired_moment.seconds - self._state.initial_seconds}:s"
        )
        # TODO: current offset is misleading here
        if event.time_signature is not None:
            new_duration = event.time_signature[0] / event.time_signature[1]
            if desired_moment.measure_offset < new_duration:
                # On the downbeat
                self._state = self._state._replace(
                    previous_measure=desired_moment.measure,
                    previous_time_signature_change_offset=(
                        desired_moment.offset - desired_moment.measure_offset
                    ),
                    time_signature=event.time_signature,
                )
            else:
                # Moving from a longer time signature to a shorter one
                # Advance to the next downbeat immediately
                self._state = self._state._replace(
                    previous_measure=desired_moment.measure + 1,
                    previous_time_signature_change_offset=desired_moment.offset,
                    time_signature=event.time_signature,
                )
            self._reschedule_measure_relative_events()
            current_moment = dataclasses.replace(
                current_moment,
                time_signature=self._state.time_signature,
                measure=self._offset_to_measure(current_moment.offset),
                measure_offset=self._offset_to_measure_offset(current_moment.offset),
            )
        if event.beats_per_minute is not None:
            # If the tempo changed, we need to revise our offset math
            self._state = self._state._replace(
                beats_per_minute=float(event.beats_per_minute),
                previous_seconds=desired_moment.seconds,
                previous_offset=desired_moment.offset,
            )
            self._reschedule_offset_relative_events()
            new_current_offset = self._seconds_to_offset(current_moment.seconds)
            logger.debug(
                f"[{self.name}] ... ... ... Revised offset from "
                f"{current_moment.offset} to {new_current_offset}"
            )
            current_moment = dataclasses.replace(
                current_moment, offset=new_current_offset
            )
            return current_moment, False
            # if new_current_offset < current_moment.offset:
            #    return current_moment, False
        return current_moment, True

    def _perform_events(self, current_moment: Moment) -> Moment:
        logger.debug(
            f"[{self.name}] ... Ready to perform at "
            f"{current_moment.seconds - self._state.initial_seconds}:s / "
            f"{current_moment.offset}:o"
        )
        while self._is_running and self._event_queue.qsize():
            (
                event,
                desired_moment,
                should_continue,
                should_break,
            ) = self._process_perform_event_loop(current_moment)
            if should_continue:
                continue
            elif should_break:
                break
            if event is None or desired_moment is None:
                raise ValueError(event, desired_moment)
            if isinstance(event, ChangeEvent):
                current_moment, should_continue = self._perform_change_event(
                    event, current_moment, desired_moment
                )
                if not should_continue:
                    break
            elif isinstance(event, CallbackEvent):
                self._perform_callback_event(event, current_moment, desired_moment)
                self._process_command_deque()
            else:
                raise ValueError(event)
        return current_moment

    def _process_command_deque(self, first_run: bool = False) -> None:
        while self._command_deque:
            logger.debug(f"[{self.name}] ... Processing command deque ({first_run})")
            command = self._command_deque.popleft()
            if self._actions_by_id.pop(command.event_id, None) is None:
                continue
            schedule_at = command.schedule_at
            logger.debug(f"[{self.name}] ... ... Scheduled at {schedule_at}")
            if command.quantization is not None:
                seconds: float
                offset: Optional[float]
                measure: Optional[int]
                # If the command was queued before the clock was started,
                # reset its reference time
                if first_run and schedule_at < self._state.initial_seconds:
                    schedule_at = self._state.initial_seconds
                seconds, offset, measure = self._get_cue_point(
                    schedule_at, command.quantization
                )
            elif command.time_unit in (TimeUnit.BEATS, TimeUnit.MEASURES):
                seconds, offset, measure = self._get_schedule_point(
                    schedule_at, command.time_unit
                )
            else:
                if first_run:
                    schedule_at += self._state.initial_seconds
                seconds, offset, measure = schedule_at, None, None
            if isinstance(command, CallbackCommand):
                event: Union[CallbackEvent, ChangeEvent] = CallbackEvent(
                    args=command.args,
                    event_id=command.event_id,
                    kwargs=command.kwargs,
                    measure=measure,
                    offset=offset,
                    procedure=command.procedure,
                    seconds=seconds,
                    event_type=command.event_type,
                    invocations=0,
                )
            elif isinstance(command, ChangeCommand):
                event = ChangeEvent(
                    beats_per_minute=command.beats_per_minute,
                    event_id=command.event_id,
                    event_type=command.event_type,
                    measure=measure,
                    offset=offset,
                    seconds=seconds,
                    time_signature=command.time_signature,
                )
            else:
                raise ValueError(command)
            self._enqueue_event(event)
            logger.debug(
                f"[{self.name}] ... ... Enqueued {type(event).__name__} "
                f"({event.event_id}) for {event.seconds}:s / {event.offset}:o"
            )

    def _reschedule_offset_relative_events(self) -> None:
        for event_id in tuple(self._offset_relative_event_ids):
            action = self._cancel(event_id)
            if action is None or not isinstance(action, Event) or action.offset is None:
                self._offset_relative_event_ids.remove(event_id)
                continue
            seconds = self._offset_to_seconds(action.offset)
            logger.debug(
                f"[{self.name}] ... ... ... Rescheduling offset-relative event "
                f"({action.event_id}) from "
                f"{action.seconds - self._state.initial_seconds}:s to "
                f"{seconds - self._state.initial_seconds}:s"
            )
            self._enqueue_event(dataclasses.replace(action, seconds=seconds))

    def _reschedule_measure_relative_events(self) -> None:
        for event_id in tuple(self._measure_relative_event_ids):
            action = self._cancel(event_id)
            if (
                action is None
                or not isinstance(action, Event)
                or action.measure is None
            ):
                self._measure_relative_event_ids.remove(event_id)
                continue
            offset = self._measure_to_offset(action.measure)
            seconds = self._offset_to_seconds(offset)
            logger.debug(
                f"[{self.name}] ... ... ... Rescheduling measure-relative event from "
                f"offset {action.offset} to {offset}"
            )
            self._enqueue_event(
                dataclasses.replace(action, offset=offset, seconds=seconds)
            )

    def _start(
        self,
        initial_time: Optional[float] = None,
        initial_offset: float = 0.0,
        initial_measure: int = 1,
        beats_per_minute: Optional[float] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ) -> None:
        if self._is_running:
            raise RuntimeError("Already started")
        if initial_time is None:
            initial_time = self.get_current_time()
        self._state = ClockState(
            beats_per_minute=beats_per_minute or self._state.beats_per_minute,
            initial_seconds=initial_time,
            previous_measure=int(initial_measure),
            previous_offset=float(initial_offset),
            previous_seconds=float(initial_time),
            previous_time_signature_change_offset=float(initial_offset),
            time_signature=time_signature or self._state.time_signature,
        )
        self._is_running = True

    def _stop(self) -> bool:
        if not self._is_running:
            return False
        self._is_running = False
        return True

    ### PUBLIC METHODS ###

    def cancel(self, event_id: int) -> Optional[Action]:
        logger.debug(f"[{self.name}] Canceling {event_id}")
        event = self._cancel(event_id)
        return event

    def change(
        self,
        beats_per_minute: Optional[float] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ) -> Optional[int]:
        if not self._is_running:
            self._state = self._state._replace(
                beats_per_minute=beats_per_minute or self._state.beats_per_minute,
                time_signature=time_signature or self._state.time_signature,
            )
            return None
        event_id = next(self._counter)
        command = ChangeCommand(
            event_id=event_id,
            event_type=EventType.CHANGE,
            beats_per_minute=beats_per_minute,
            time_signature=time_signature,
            quantization=None,
            schedule_at=self.get_current_time(),
            time_unit=None,
        )
        self._enqueue_command(command)
        return event_id

    def cue(
        self,
        procedure: Callable,
        *,
        args: Optional[Tuple[Any, ...]] = None,
        event_type: int = EventType.SCHEDULE,
        kwargs: Optional[Dict[str, Any]] = None,
        quantization: Optional[Quantization] = None,
    ) -> int:
        if event_type <= 0:
            raise ValueError(f"Invalid event type {event_type}")
        elif quantization is not None and quantization not in self._valid_quantizations:
            raise ValueError(f"Invalid quantization: {quantization}")
        event_id = next(self._counter)
        command = CallbackCommand(
            args=args,
            event_id=event_id,
            event_type=event_type,
            kwargs=kwargs,
            procedure=procedure,
            quantization=quantization,
            schedule_at=self.get_current_time() if self.is_running else 0,
            time_unit=None,
        )
        self._enqueue_command(command)
        return event_id

    def cue_change(
        self,
        *,
        beats_per_minute: Optional[float] = None,
        quantization: Optional[Quantization] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ) -> int:
        if quantization is not None and quantization not in self._valid_quantizations:
            raise ValueError(f"Invalid quantization: {quantization}")
        event_id = next(self._counter)
        command = ChangeCommand(
            beats_per_minute=beats_per_minute,
            event_id=event_id,
            event_type=EventType.CHANGE,
            quantization=quantization,
            schedule_at=self.get_current_time() if self.is_running else 0,
            time_signature=time_signature,
            time_unit=None,
        )
        self._enqueue_command(command)
        return event_id

    def get_current_time(self) -> float:
        return time.time()

    def peek(self) -> Optional[Event]:
        try:
            return self._event_queue.peek()
        except queue.Empty:
            return None

    @classmethod
    def quantization_to_beats(cls, quantization: Quantization) -> float:
        fraction = fractions.Fraction(quantization.replace("T", ""))
        if "T" in quantization:
            fraction *= fractions.Fraction(2, 3)
        return float(fraction)

    def reschedule(
        self,
        event_id: int,
        *,
        schedule_at: float = 0.0,
        time_unit: TimeUnit = TimeUnit.BEATS,
    ) -> Optional[int]:
        if (event_or_command := self.cancel(event_id)) is None:
            return None
        if isinstance(event_or_command, Command):
            command: Command = dataclasses.replace(
                event_or_command, schedule_at=schedule_at, time_unit=time_unit
            )
        elif isinstance(event_or_command, CallbackEvent):
            command = CallbackCommand(
                args=event_or_command.args,
                event_id=event_or_command.event_id,
                event_type=event_or_command.event_type,
                kwargs=event_or_command.kwargs,
                procedure=event_or_command.procedure,
                quantization=None,
                schedule_at=schedule_at,
                time_unit=time_unit,
            )
        elif isinstance(event_or_command, ChangeEvent):
            command = ChangeCommand(
                beats_per_minute=event_or_command.beats_per_minute,
                event_id=event_or_command.event_id,
                event_type=EventType.CHANGE,
                quantization=None,
                schedule_at=schedule_at,
                time_signature=event_or_command.time_signature,
                time_unit=time_unit,
            )
        else:
            raise ValueError(event_or_command)
        self._enqueue_command(command)
        return event_id

    def schedule(
        self,
        procedure: Callable,
        *,
        event_type: int = EventType.SCHEDULE,
        schedule_at: float = 0.0,
        time_unit: TimeUnit = TimeUnit.BEATS,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> int:
        logger.debug(f"[{self.name}] Scheduling {procedure}")
        if event_type <= 0:
            raise ValueError(f"Invalid event type {event_type}")
        event_id = next(self._counter)
        command = CallbackCommand(
            args=args,
            event_id=event_id,
            event_type=event_type,
            kwargs=kwargs,
            procedure=procedure,
            quantization=None,
            schedule_at=schedule_at,
            time_unit=time_unit,
        )
        self._enqueue_command(command)
        return event_id

    def schedule_change(
        self,
        *,
        beats_per_minute: Optional[float] = None,
        schedule_at: float = 0.0,
        time_signature: Optional[Tuple[int, int]] = None,
        time_unit: TimeUnit = TimeUnit.BEATS,
    ) -> int:
        event_id = next(self._counter)
        command = ChangeCommand(
            beats_per_minute=beats_per_minute,
            event_id=event_id,
            event_type=EventType.CHANGE,
            quantization=None,
            schedule_at=schedule_at,
            time_signature=time_signature,
            time_unit=time_unit,
        )
        self._enqueue_command(command)
        return event_id

    ### PUBLIC PROPERTIES ###

    @property
    def beats_per_minute(self) -> float:
        return self._state.beats_per_minute

    @property
    def is_running(self) -> bool:
        return self._is_running

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def slop(self) -> float:
        return self._slop

    @slop.setter
    def slop(self, slop: float) -> None:
        if slop <= 0:
            raise ValueError(slop)
        self._slop = float(slop)

    @property
    def time_signature(self) -> Tuple[int, int]:
        return self._state.time_signature



# flake8: noqa
import os
import platform

import pytest
from uqbar.strings import normalize

from supriya.ugens import (
    FFT,
    IFFT,
    LocalBuf,
    Out,
    PV_BinScramble,
    PV_MagFreeze,
    PV_MagMul,
    PinkNoise,
    SuperColliderSynthDef,
    SynthDefBuilder,
    decompile_synthdef,
)


def test_01():
    with SynthDefBuilder() as builder:
        local_buf = LocalBuf.ir(frame_count=2048)
        source = PinkNoise.ar()
        pv_chain = FFT.kr(buffer_id=local_buf, source=source)
        ifft = IFFT.ar(pv_chain=pv_chain)
        Out.ar(bus=0, source=ifft)
    py_synthdef = builder.build("LocalBufTest")
    # fmt: off
    test_compiled_synthdef = (
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x0cLocalBufTest'
                b'\x00\x00\x00\x04'
                    b'?\x80\x00\x00'
                    b'E\x00\x00\x00'
                    b'?\x00\x00\x00'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x06'
                    b'\x0cMaxLocalBufs'
                        b'\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\x00'
                    b'\x08LocalBuf'
                        b'\x00'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00'
                    b'\tPinkNoise'
                        b'\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x02'
                    b'\x03FFT'
                        b'\x01'
                        b'\x00\x00\x00\x06'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x02'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x02'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x03'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x03'
                            b'\x01'
                    b'\x04IFFT'
                        b'\x02'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x03'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x03'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x03'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x03'
                            b'\x00\x00\x00\x04'
                            b'\x00\x00\x00\x00'
                b'\x00\x00'
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.fixture
def py_synthdef_02():
    with SynthDefBuilder() as builder:
        source = PinkNoise.ar()
        local_buf = LocalBuf.ir(frame_count=2048)
        pv_chain = FFT.kr(buffer_id=local_buf, source=source)
        pv_chain_a = PV_BinScramble.kr(pv_chain=pv_chain)
        pv_chain_b = PV_MagFreeze.kr(pv_chain=pv_chain)
        pv_chain = PV_MagMul.kr(pv_chain_a=pv_chain_a, pv_chain_b=pv_chain_b)
        ifft = IFFT.ar(pv_chain=pv_chain)
        Out.ar(bus=0, source=ifft)
    py_synthdef = builder.build("PVCopyTest")
    return py_synthdef


def test_02_ugens(py_synthdef_02):
    assert tuple(repr(_) for _ in py_synthdef_02.ugens) == (
        "<PinkNoise.ar()>",
        "<MaxLocalBufs.ir()>",
        "<LocalBuf.ir()>",
        "<FFT.kr()>",
        "<BufFrames.ir()>",
        "<LocalBuf.ir()>",
        "<PV_Copy.kr()>",
        "<PV_BinScramble.kr()>",
        "<PV_MagFreeze.kr()>",
        "<PV_MagMul.kr()>",
        "<IFFT.ar()>",
        "<Out.ar()>",
    )
    assert str(py_synthdef_02) == normalize(
        """
        synthdef:
            name: PVCopyTest
            ugens:
            -   PinkNoise.ar: null
            -   MaxLocalBufs.ir:
                    maximum: 2.0
            -   LocalBuf.ir/0:
                    channel_count: 1.0
                    frame_count: 2048.0
            -   FFT.kr:
                    buffer_id: LocalBuf.ir/0[0]
                    source: PinkNoise.ar[0]
                    hop: 0.5
                    window_type: 0.0
                    active: 1.0
                    window_size: 0.0
            -   BufFrames.ir:
                    buffer_id: LocalBuf.ir/0[0]
            -   LocalBuf.ir/1:
                    channel_count: 1.0
                    frame_count: BufFrames.ir[0]
            -   PV_Copy.kr:
                    pv_chain_a: FFT.kr[0]
                    pv_chain_b: LocalBuf.ir/1[0]
            -   PV_BinScramble.kr:
                    pv_chain: PV_Copy.kr[0]
                    wipe: 0.0
                    width: 0.2
                    trigger: 0.0
            -   PV_MagFreeze.kr:
                    pv_chain: FFT.kr[0]
                    freeze: 0.0
            -   PV_MagMul.kr:
                    pv_chain_a: PV_BinScramble.kr[0]
                    pv_chain_b: PV_MagFreeze.kr[0]
            -   IFFT.ar:
                    pv_chain: PV_MagMul.kr[0]
                    window_type: 0.0
                    window_size: 0.0
            -   Out.ar:
                    bus: 0.0
                    source[0]: IFFT.ar[0]
        """
    )


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_02_supriya_vs_sclang(py_synthdef_02):
    sc_synthdef = SuperColliderSynthDef(
        "PVCopyTest",
        r"""
        var source, pv_chain, pv_chain_a, pv_chain_b, ifft, out;
        source = PinkNoise.ar();
        pv_chain = FFT(LocalBuf(2048), source);
        pv_chain_a = PV_BinScramble(pv_chain);
        pv_chain_b = PV_MagFreeze(pv_chain);
        pv_chain = PV_MagMul(pv_chain_a, pv_chain_b);
        ifft = IFFT.ar(pv_chain);
        out = Out.ar(0, ifft);
        """,
    )
    sc_compiled_synthdef = bytes(sc_synthdef.compile())
    py_compiled_synthdef = py_synthdef_02.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef
    sc_synthdef = decompile_synthdef(sc_compiled_synthdef)
    assert tuple(repr(_) for _ in sc_synthdef.ugens) == (
        "<PinkNoise.ar()>",
        "<MaxLocalBufs.ir()>",
        "<LocalBuf.ir()>",
        "<FFT.kr()>",
        "<BufFrames.ir()>",
        "<LocalBuf.ir()>",
        "<PV_Copy.kr()>",
        "<PV_BinScramble.kr()>",
        "<PV_MagFreeze.kr()>",
        "<PV_MagMul.kr()>",
        "<IFFT.ar()>",
        "<Out.ar()>",
    )
    assert tuple(repr(_) for _ in sc_synthdef.ugens) == tuple(
        repr(_) for _ in py_synthdef_02.ugens
    )


def test_02_supriya_vs_bytes(py_synthdef_02):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\nPVCopyTest'
                b'\x00\x00\x00\x06'
                    b'@\x00\x00\x00'
                    b'?\x80\x00\x00'
                    b'E\x00\x00\x00'
                    b'?\x00\x00\x00'
                    b'\x00\x00\x00\x00'
                    b'>L\xcc\xcd'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x0c'
                    b'\tPinkNoise'
                        b'\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x02'
                    b'\x0cMaxLocalBufs'
                        b'\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\x00'
                    b'\x08LocalBuf'
                        b'\x00'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x01'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x02'
                            b'\x00\x00\x00\x01'
                                b'\x00\x00\x00\x00'
                            b'\x00'
                    b'\x03FFT'
                        b'\x01'
                        b'\x00\x00\x00\x06'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x02'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x03'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x04'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x01'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x04'
                            b'\x01'
                    b'\tBufFrames'
                        b'\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x02'
                                b'\x00\x00\x00\x00'
                            b'\x00'
                    b'\x08LocalBuf'
                        b'\x00'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x04'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x01'
                                b'\x00\x00\x00\x00'
                            b'\x00'
                    b'\x07PV_Copy'
                        b'\x01'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x03'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x05'
                                b'\x00\x00\x00\x00'
                            b'\x01'
                    b'\x0ePV_BinScramble'
                        b'\x01'
                        b'\x00\x00\x00\x04'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x06'
                                b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x04'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x05'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x04'
                            b'\x01'
                    b'\x0cPV_MagFreeze'
                        b'\x01'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x03'
                                b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x04'
                            b'\x01'
                    b'\tPV_MagMul'
                        b'\x01'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x07'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x08'
                                b'\x00\x00\x00\x00'
                            b'\x01'
                    b'\x04IFFT'
                        b'\x02'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\t'
                                b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x04'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x04'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff\x00\x00\x00\x04\x00\x00\x00\n\x00\x00\x00\x00'
                b'\x00\x00'
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_02.compile()
    assert py_compiled_synthdef == test_compiled_synthdef



import os
import platform

import pytest
from uqbar.strings import normalize

from supriya.ugens import SuperColliderSynthDef, decompile_synthdef


@pytest.fixture
def sc_synthdef_expansion():
    return SuperColliderSynthDef(
        "expansionTest",
        """
        var a, b, c, d;
        a = SinOsc.ar([1, 2]);
        b = Pan2.ar(a);
        c = Pan2.ar(b).softclip;
        Out.ar(0, c);
        """,
    )


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_sc_format(sc_synthdef_expansion):
    sc_compiled_synthdef = bytes(sc_synthdef_expansion.compile())
    sc_synthdef = decompile_synthdef(sc_compiled_synthdef)
    assert str(sc_synthdef) == normalize(
        """
        synthdef:
            name: expansionTest
            ugens:
            -   SinOsc.ar/0:
                    frequency: 1.0
                    phase: 0.0
            -   Pan2.ar/0:
                    source: SinOsc.ar/0[0]
                    position: 0.0
                    level: 1.0
            -   Pan2.ar/1:
                    source: Pan2.ar/0[0]
                    position: 0.0
                    level: 1.0
            -   UnaryOpUGen(SOFTCLIP).ar/0:
                    source: Pan2.ar/1[0]
            -   UnaryOpUGen(SOFTCLIP).ar/1:
                    source: Pan2.ar/1[1]
            -   Pan2.ar/2:
                    source: Pan2.ar/0[1]
                    position: 0.0
                    level: 1.0
            -   UnaryOpUGen(SOFTCLIP).ar/2:
                    source: Pan2.ar/2[0]
            -   UnaryOpUGen(SOFTCLIP).ar/3:
                    source: Pan2.ar/2[1]
            -   SinOsc.ar/1:
                    frequency: 2.0
                    phase: 0.0
            -   Pan2.ar/3:
                    source: SinOsc.ar/1[0]
                    position: 0.0
                    level: 1.0
            -   Pan2.ar/4:
                    source: Pan2.ar/3[0]
                    position: 0.0
                    level: 1.0
            -   UnaryOpUGen(SOFTCLIP).ar/4:
                    source: Pan2.ar/4[0]
            -   Out.ar/0:
                    bus: 0.0
                    source[0]: UnaryOpUGen(SOFTCLIP).ar/0[0]
                    source[1]: UnaryOpUGen(SOFTCLIP).ar/4[0]
            -   UnaryOpUGen(SOFTCLIP).ar/5:
                    source: Pan2.ar/4[1]
            -   Out.ar/1:
                    bus: 0.0
                    source[0]: UnaryOpUGen(SOFTCLIP).ar/1[0]
                    source[1]: UnaryOpUGen(SOFTCLIP).ar/5[0]
            -   Pan2.ar/5:
                    source: Pan2.ar/3[1]
                    position: 0.0
                    level: 1.0
            -   UnaryOpUGen(SOFTCLIP).ar/6:
                    source: Pan2.ar/5[0]
            -   Out.ar/2:
                    bus: 0.0
                    source[0]: UnaryOpUGen(SOFTCLIP).ar/2[0]
                    source[1]: UnaryOpUGen(SOFTCLIP).ar/6[0]
            -   UnaryOpUGen(SOFTCLIP).ar/7:
                    source: Pan2.ar/5[1]
            -   Out.ar/3:
                    bus: 0.0
                    source[0]: UnaryOpUGen(SOFTCLIP).ar/3[0]
                    source[1]: UnaryOpUGen(SOFTCLIP).ar/7[0]
        """
    )



import uuid
from unittest.mock import Mock, call

import pytest

from supriya.contexts import BusGroup, Server
from supriya.enums import CalculationRate
from supriya.patterns.events import BusAllocateEvent, Priority

id_ = uuid.uuid4()


@pytest.mark.parametrize(
    "event, offset, expected",
    [
        (
            BusAllocateEvent(id_, calculation_rate="control", channel_count=2),
            0.0,
            [
                (
                    0.0,
                    Priority.START,
                    BusAllocateEvent(id_, calculation_rate="control", channel_count=2),
                )
            ],
        ),
        (
            BusAllocateEvent(id_, calculation_rate="audio", channel_count=8),
            0.0,
            [
                (
                    0.0,
                    Priority.START,
                    BusAllocateEvent(id_, calculation_rate="audio", channel_count=8),
                )
            ],
        ),
    ],
)
def test_expand(event, offset, expected):
    actual = event.expand(offset)
    assert actual == expected


def test_perform():
    context = Server().boot()
    spy = Mock(wraps=context)
    proxy_mapping = {}
    notes_mapping = {}
    # Allocate
    event = BusAllocateEvent(id_, calculation_rate="audio", channel_count=8)
    with context.at():
        event.perform(
            spy,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    assert proxy_mapping == {
        id_: BusGroup(
            id_=16,
            context=context,
            count=8,
            calculation_rate=CalculationRate.AUDIO,
        )
    }
    assert notes_mapping == {}
    assert spy.mock_calls == [
        call.add_bus_group(calculation_rate=CalculationRate.AUDIO, count=8)
    ]



import logging
from pathlib import Path

import pytest

from supriya import assets
from supriya.contexts.errors import MomentClosed
from supriya.contexts.nonrealtime import Score
from supriya.osc import OscBundle, OscMessage


@pytest.fixture
def audio_paths():
    return sorted((Path(assets.__path__[0]) / "audio" / "birds").glob("*.wav"))


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest.fixture
def context():
    return Score()


def test_add_buffer(audio_paths, context):
    with context.at(0):
        # neither frame count nor file path provided
        with pytest.raises(ValueError):
            context.add_buffer()
        # both channel count and file path provided
        with pytest.raises(ValueError):
            context.add_buffer(file_path=audio_paths[0], channel_count=1)
        # both channel count and channel indices provided
        with pytest.raises(ValueError):
            context.add_buffer(
                file_path=audio_paths[0], channel_count=1, channel_indices=[0]
            )
        # /b_alloc
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        # /b_allocRead
        buffer_b = context.add_buffer(frame_count=19, file_path=audio_paths[0])
        # /b_allocReadChannel
        buffer_c = context.add_buffer(
            frame_count=17, file_path=audio_paths[1], channel_indices=[1, 0]
        )
        # completion without moment errors
        buffer_d = context.add_buffer(channel_count=3, frame_count=31)
    with pytest.raises(MomentClosed):
        with buffer_d:
            ...
    # completion without moment via on_completion lambda succeeds
    with context.at(0):
        buffer_e = context.add_buffer(
            channel_count=1, frame_count=47, on_completion=lambda ctx: ctx.add_group()
        )
    # completion inside moment succeeds
    with context.at(1.23):
        buffer_f = context.add_buffer(channel_count=3, frame_count=31)
        with buffer_f:
            context.add_group()
    buffers = [buffer_a, buffer_b, buffer_c, buffer_d, buffer_e, buffer_f]
    assert all(buffer.context is context for buffer in buffers)
    assert [buffer.id_ for buffer in buffers] == [0, 1, 2, 3, 4, 5]
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_allocRead", 1, str(audio_paths[0]), 0, 19),
                OscMessage("/b_allocReadChannel", 2, str(audio_paths[1]), 0, 17, 1, 0),
                OscMessage("/b_alloc", 3, 31, 3),
                OscMessage("/b_alloc", 4, 47, 1, OscMessage("/g_new", 1000, 0, 0)),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 5, 31, 3, OscMessage("/g_new", 1001, 0, 0)),
            ),
            timestamp=1.23,
        ),
    ]


def test_add_buffer_group(context):
    with context.at(0):
        # neither channel count nor frame count provided
        with pytest.raises(ValueError):
            context.add_buffer_group()
        # count less than 1
        with pytest.raises(ValueError):
            context.add_buffer_group(channel_count=1, count=0, frame_count=512)
        # /b_alloc
        buffer_group = context.add_buffer_group(
            channel_count=1, count=5, frame_count=23
        )
    assert len(buffer_group) == 5
    assert all(buffer.context is context for buffer in buffer_group)
    assert [buffer.id_ for buffer in buffer_group] == [0, 1, 2, 3, 4]
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_alloc", 2, 23, 1),
                OscMessage("/b_alloc", 3, 23, 1),
                OscMessage("/b_alloc", 4, 23, 1),
            ),
            timestamp=0.0,
        )
    ]


def test_close_buffer(context):
    with context.at(0):
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        buffer_b = context.add_buffer(channel_count=1, frame_count=23)
        buffer_c = context.add_buffer(channel_count=1, frame_count=23)
        # completion without moment errors, but initial request succeeds
        completion = buffer_a.close()
    with pytest.raises(MomentClosed):
        with completion:
            ...
    # completion via on_completion lambda succeeds
    with context.at(0):
        buffer_b.close(on_completion=lambda ctx: buffer_b.free())
    # completion inside moment succeeds
    with context.at(0):
        with buffer_c.close():
            buffer_c.free()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_alloc", 2, 23, 1),
                OscMessage("/b_close", 0),
                OscMessage("/b_close", 1, OscMessage("/b_free", 1)),
                OscMessage("/b_close", 2, OscMessage("/b_free", 2)),
            ),
            timestamp=0.0,
        )
    ]


def test_copy_buffer(context):
    with context.at(0):
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        buffer_b = context.add_buffer(channel_count=1, frame_count=23)
        buffer_a.copy(
            target_buffer=buffer_b,
            starting_frame=5,
            target_starting_frame=7,
            frame_count=3,
        )
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_gen", 1, "copy", 7, 0, 5, 3),
            ),
            timestamp=0.0,
        )
    ]


def test_fill_buffer(context):
    with context.at(0):
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        buffer_b = context.add_buffer(channel_count=1, frame_count=23)
        # one-off
        buffer_a.fill(2, 3, 0.5)
    with context.at(1.23):
        # with moment, including grouping
        buffer_a.fill(2, 3, 0.6)
        buffer_b.fill(3, 4, 0.5)
        buffer_a.fill(4, 5, 0.4)
        buffer_a.fill(5, 6, 0.3)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_fill", 0, 2, 3, 0.5),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage("/b_fill", 0, 2, 3, 0.6, 4, 5, 0.4, 5, 6, 0.3),
                OscMessage("/b_fill", 1, 3, 4, 0.5),
            ),
            timestamp=1.23,
        ),
    ]


def test_free_buffer(context):
    with context.at(0):
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        buffer_b = context.add_buffer(channel_count=1, frame_count=23)
        buffer_c = context.add_buffer(channel_count=1, frame_count=23)
        completion = buffer_a.free()
    with pytest.raises(MomentClosed):
        # completion without moment errors, but initial request succeeds
        with completion:
            ...
    with context.at(1.23):
        # completion via on_completion lambda succeeds
        buffer_b.free(on_completion=lambda ctx: ctx.add_group())
        # completion inside moment succeeds
        with buffer_c.free():
            context.add_group()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_alloc", 2, 23, 1),
                OscMessage("/b_free", 0),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage("/b_free", 1, OscMessage("/g_new", 1000, 0, 0)),
                OscMessage("/b_free", 2, OscMessage("/g_new", 1001, 0, 0)),
            ),
            timestamp=1.23,
        ),
    ]


def test_generate_buffer(context):
    with context.at(0):
        buffer = context.add_buffer(channel_count=1, frame_count=1024)
        buffer.generate(command_name="sine1", amplitudes=[1, 2, 3])
        buffer.generate(
            command_name="sine2", amplitudes=[1, 2, 3], frequencies=[4, 5, 6]
        )
        buffer.generate(
            command_name="sine3",
            amplitudes=[1, 2, 3],
            frequencies=[4, 5, 6],
            phases=[0.25, 0.0, 0.5],
        )
        buffer.generate(command_name="cheby", amplitudes=[1, 2, 3])
        buffer.generate(command_name="sine1", amplitudes=[1, 2, 3], as_wavetable=True)
        buffer.generate(
            command_name="sine1", amplitudes=[1, 2, 3], should_clear_first=True
        )
        buffer.generate(
            command_name="sine1",
            amplitudes=[1, 2, 3],
            should_clear_first=True,
            should_normalize=True,
        )
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 1024, 1),
                OscMessage("/b_gen", 0, "sine1", 0, 1.0, 2.0, 3.0),
                OscMessage("/b_gen", 0, "sine2", 0, 4.0, 1.0, 5.0, 2.0, 6.0, 3.0),
                OscMessage(
                    "/b_gen",
                    0,
                    "sine3",
                    0,
                    4.0,
                    1.0,
                    0.25,
                    5.0,
                    2.0,
                    0.0,
                    6.0,
                    3.0,
                    0.5,
                ),
                OscMessage("/b_gen", 0, "cheby", 0, 1.0, 2.0, 3.0),
                OscMessage("/b_gen", 0, "sine1", 2, 1.0, 2.0, 3.0),
                OscMessage("/b_gen", 0, "sine1", 4, 1.0, 2.0, 3.0),
                OscMessage("/b_gen", 0, "sine1", 5, 1.0, 2.0, 3.0),
            ),
            timestamp=0.0,
        ),
    ]


def test_normalize_buffer(context):
    with context.at(0):
        buffer = context.add_buffer(channel_count=1, frame_count=23)
        buffer.normalize(0.5)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_gen", 0, "normalize", 0.5),
            ),
            timestamp=0.0,
        )
    ]


def test_read_buffer(audio_paths, context):
    with context.at(0):
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        buffer_b = context.add_buffer(channel_count=1, frame_count=23)
        buffer_c = context.add_buffer(channel_count=1, frame_count=23)
        # completion without moment errors, but initial request succeeds
        completion = buffer_a.read(file_path=audio_paths[0])
    with pytest.raises(MomentClosed):
        with completion:
            ...
    with context.at(1.23):
        # completion via on_completion lambda succeeds
        buffer_b.read(
            file_path=audio_paths[1], on_completion=lambda ctx: ctx.add_group()
        )
        # completion inside moment succeeds
        with buffer_c.read(file_path=audio_paths[2]):
            context.add_group()
        # parameters
        buffer_a.read(
            buffer_starting_frame=5,
            channel_indices=[0],
            file_path=audio_paths[3],
            frame_count=16,
            leave_open=True,
            starting_frame=3,
        )
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_alloc", 2, 23, 1),
                OscMessage("/b_read", 0, str(audio_paths[0]), 0, -1, 0, 0),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage(
                    "/b_read",
                    1,
                    str(audio_paths[1]),
                    0,
                    -1,
                    0,
                    0,
                    OscMessage("/g_new", 1000, 0, 0),
                ),
                OscMessage(
                    "/b_read",
                    2,
                    str(audio_paths[2]),
                    0,
                    -1,
                    0,
                    0,
                    OscMessage("/g_new", 1001, 0, 0),
                ),
                OscMessage("/b_readChannel", 0, str(audio_paths[3]), 3, 16, 5, 1, 0),
            ),
            timestamp=1.23,
        ),
    ]


def test_set_buffer(context):
    with context.at(0):
        buffer = context.add_buffer(channel_count=1, frame_count=32)
        buffer.set(2, 0.5)
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 32, 1),
                OscMessage("/b_set", 0, 2, 0.5),
            ),
            timestamp=0.0,
        )
    ]


def test_set_buffer_range(context):
    with context.at(0):
        buffer = context.add_buffer(channel_count=1, frame_count=32)
        buffer.set_range(4, (0.5, 0.75, 0.25))
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 32, 1),
                OscMessage("/b_setn", 0, 4, 3, 0.5, 0.75, 0.25),
            ),
            timestamp=0.0,
        )
    ]


def test_write_buffer(context, tmp_path):
    with context.at(0):
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        buffer_b = context.add_buffer(channel_count=1, frame_count=23)
        buffer_c = context.add_buffer(channel_count=1, frame_count=23)
        # completion without moment errors, but initial request succeeds
        completion = buffer_a.write(file_path=tmp_path / "foo-1.aiff")
    with pytest.raises(MomentClosed):
        with completion:
            ...
    with context.at(1.23):
        # completion via on_completion lambda succeeds
        buffer_b.write(
            file_path=tmp_path / "foo-2.aiff", on_completion=lambda ctx: ctx.add_group()
        )
        # completion inside moment succeeds
        with buffer_c.write(file_path=tmp_path / "foo-3.aiff"):
            context.add_group()
        # parameters
        buffer_a.write(
            file_path=tmp_path / "foo-4.wav",
            frame_count=16,
            header_format="WAV",
            leave_open=True,
            sample_format="FLOAT",
            starting_frame=5,
        )
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_alloc", 2, 23, 1),
                OscMessage(
                    "/b_write",
                    0,
                    str(tmp_path / "foo-1.aiff"),
                    "aiff",
                    "int24",
                    -1,
                    0,
                    0,
                ),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage(
                    "/b_write",
                    1,
                    str(tmp_path / "foo-2.aiff"),
                    "aiff",
                    "int24",
                    -1,
                    0,
                    0,
                    OscMessage("/g_new", 1000, 0, 0),
                ),
                OscMessage(
                    "/b_write",
                    2,
                    str(tmp_path / "foo-3.aiff"),
                    "aiff",
                    "int24",
                    -1,
                    0,
                    0,
                    OscMessage("/g_new", 1001, 0, 0),
                ),
                OscMessage(
                    "/b_write", 0, str(tmp_path / "foo-4.wav"), "wav", "float", 16, 5, 1
                ),
            ),
            timestamp=1.23,
        ),
    ]


def test_zero_buffer(context):
    with context.at(0):
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        buffer_b = context.add_buffer(channel_count=1, frame_count=23)
        buffer_c = context.add_buffer(channel_count=1, frame_count=23)
        # completion without moment errors
        completion = buffer_a.zero()
    with pytest.raises(MomentClosed):
        with completion:
            ...
    with context.at(1.23):
        # completion via on_completion lambda succeeds
        buffer_b.zero(on_completion=lambda ctx: ctx.add_group())
        # completion inside moment succeeds
        with buffer_c.zero():
            context.add_group()
    assert list(context.iterate_osc_bundles()) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_alloc", 2, 23, 1),
                OscMessage("/b_zero", 0),
            ),
            timestamp=0.0,
        ),
        OscBundle(
            contents=(
                OscMessage("/b_zero", 1, OscMessage("/g_new", 1000, 0, 0)),
                OscMessage("/b_zero", 2, OscMessage("/g_new", 1001, 0, 0)),
            ),
            timestamp=1.23,
        ),
    ]



from ...ugens import Out, SinOsc, SynthDef, SynthDefBuilder


def _build_synthdef() -> SynthDef:
    with SynthDefBuilder(amplitude=0, bus=0, frequency=440) as builder:
        Out.ar(
            bus=builder["bus"],
            source=SinOsc.ar(frequency=builder["frequency"]) * builder["amplitude"],
        )
    return builder.build(name="simple_sine")


simple_sine = _build_synthdef()

__all__ = ["simple_sine"]



import uuid
from unittest.mock import Mock, call

import pytest

from supriya import AddAction
from supriya.assets.synthdefs import default
from supriya.contexts import Server, Synth
from supriya.patterns.events import Priority, SynthAllocateEvent

id_ = uuid.uuid4()


@pytest.mark.parametrize(
    "event, offset, expected",
    [
        (
            SynthAllocateEvent(id_, default),
            0.0,
            [(0.0, Priority.START, SynthAllocateEvent(id_, default))],
        )
    ],
)
def test_expand(event, offset, expected):
    actual = event.expand(offset)
    assert actual == expected


def test_perform():
    context = Server().boot()
    spy = Mock(wraps=context)
    proxy_mapping = {}
    notes_mapping = {}
    # Allocate
    event = SynthAllocateEvent(id_, default)
    with context.at():
        event.perform(
            spy,
            proxy_mapping,
            current_offset=0.0,
            notes_mapping=notes_mapping,
            priority=Priority.START,
        )
    assert proxy_mapping == {id_: Synth(context=context, id_=1000, synthdef=default)}
    assert notes_mapping == {}
    assert spy.mock_calls == [
        call.add_synth(
            add_action=AddAction.ADD_TO_HEAD,
            permanent=False,
            synthdef=default,
            target_node=None,
        )
    ]



import asyncio
import logging

import pytest
import pytest_asyncio

from supriya import default
from supriya.contexts.errors import InvalidCalculationRate
from supriya.contexts.realtime import AsyncServer, Server
from supriya.osc import OscMessage


async def get(x):
    if asyncio.iscoroutine(x):
        return await x
    return x


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest_asyncio.fixture(autouse=True, params=[AsyncServer, Server])
async def context(request):
    context = request.param()
    await get(context.boot())
    context.add_synthdefs(default)
    await get(context.sync())
    yield context


@pytest.mark.asyncio
async def test_Bus_allocated(context):
    # TODO: what are the semantics actually?
    #       buses always exist.
    #       but the /leasing/ of a bus ID is temporal.
    #       what we need is a distinction between the permanent...
    #       ... and the leased.
    audio_bus = context.add_bus("AUDIO")
    control_bus = context.add_bus("CONTROL")
    assert audio_bus.allocated
    assert control_bus.allocated
    audio_bus.free()
    control_bus.free()
    assert audio_bus.allocated
    assert control_bus.allocated
    await get(context.quit())
    assert not audio_bus.allocated
    assert not control_bus.allocated


@pytest.mark.asyncio
async def test_add_bus(context):
    # invalid calculation rate
    with pytest.raises(InvalidCalculationRate):
        context.add_bus("SCALAR")
    # ok
    with context.osc_protocol.capture() as transcript:
        audio_bus = context.add_bus("AUDIO")
        control_bus = context.add_bus("CONTROL")
    assert audio_bus.context is context
    assert audio_bus.id_ == 16
    assert control_bus.context is context
    assert control_bus.id_ == 0
    assert transcript.filtered(received=False, status=False) == []


@pytest.mark.asyncio
async def test_add_bus_group(context):
    # invalid calculation rate
    with pytest.raises(InvalidCalculationRate):
        context.add_bus_group("SCALAR")
    # count less than 1
    with pytest.raises(ValueError):
        context.add_bus_group("AUDIO", 0)
    # ok
    with context.osc_protocol.capture() as transcript:
        audio_bus_group = context.add_bus_group("AUDIO", 8)
        control_bus_group = context.add_bus_group("CONTROL", 4)
    assert len(audio_bus_group) == 8
    assert all(audio_bus.context is context for audio_bus in audio_bus_group)
    assert [audio_bus.id_ for audio_bus in audio_bus_group] == [
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
    ]
    assert len(control_bus_group) == 4
    assert all(control_bus.context is context for control_bus in control_bus_group)
    assert [control_bus.id_ for control_bus in control_bus_group] == [0, 1, 2, 3]
    assert transcript.filtered(received=False, status=False) == []


@pytest.mark.asyncio
async def test_fill_buses(context):
    audio_bus = context.add_bus("AUDIO")
    control_bus_a = context.add_bus("CONTROL")
    control_bus_b = context.add_bus("CONTROL")
    control_bus_c = context.add_bus("CONTROL")
    with context.osc_protocol.capture() as transcript:
        with pytest.raises(InvalidCalculationRate):
            audio_bus.fill(2, 0.75)
        control_bus_a.fill(3, 0.5)
        with context.at():
            control_bus_b.fill(4, 0.25)
            control_bus_c.fill(5, 0.125)
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/c_fill", 0, 3, 0.5),
        OscMessage("/c_fill", 1, 4, 0.25, 2, 5, 0.125),
    ]


@pytest.mark.asyncio
async def test_free_bus(context):
    audio_bus = context.add_bus("AUDIO")
    control_bus = context.add_bus("CONTROL")
    with context.osc_protocol.capture() as transcript:
        audio_bus.free()
        control_bus.free()
    assert transcript.filtered(received=False, status=False) == []
    new_audio_bus = context.add_bus("AUDIO")
    new_control_bus = context.add_bus("CONTROL")
    # verify IDs are re-used
    assert audio_bus.id_ == new_audio_bus.id_
    assert control_bus.id_ == new_control_bus.id_


@pytest.mark.asyncio
async def test_get_bus(context):
    audio_bus = context.add_bus("audio")
    with pytest.raises(InvalidCalculationRate):
        await get(audio_bus.get())
    bus = context.add_bus()
    assert await get(bus.get()) == 0.0
    # unsync
    with context.osc_protocol.capture() as transcript:
        assert await get(bus.get(sync=False)) is None
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/c_get", 0)
    ]


@pytest.mark.asyncio
async def test_get_bus_range(context):
    audio_bus_group = context.add_bus_group("audio", count=4)
    with pytest.raises(InvalidCalculationRate):
        await get(audio_bus_group[0].get_range(count=len(audio_bus_group)))
    bus_group = context.add_bus_group(count=4)
    assert await get(bus_group[0].get_range(count=len(bus_group))) == (
        0.0,
        0.0,
        0.0,
        0.0,
    )
    with context.osc_protocol.capture() as transcript:
        assert (
            await get(bus_group[0].get_range(count=len(bus_group), sync=False)) is None
        )
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/c_getn", 0, 4)
    ]


@pytest.mark.asyncio
async def test_set_bus(context):
    audio_bus = context.add_bus("AUDIO")
    control_bus_a = context.add_bus("CONTROL")
    control_bus_b = context.add_bus("CONTROL")
    control_bus_c = context.add_bus("CONTROL")
    with context.osc_protocol.capture() as transcript:
        with pytest.raises(InvalidCalculationRate):
            audio_bus.set(0.75)
        control_bus_a.set(0.5)
        with context.at():
            control_bus_b.set(0.25)
            control_bus_c.set(0.125)
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/c_set", 0, 0.5),
        OscMessage("/c_set", 1, 0.25, 2, 0.125),
    ]


@pytest.mark.asyncio
async def test_set_bus_range(context):
    audio_bus_group = context.add_bus_group("AUDIO", count=4)
    control_bus_group = context.add_bus_group("CONTROL", count=4)
    with context.osc_protocol.capture() as transcript:
        with pytest.raises(InvalidCalculationRate):
            audio_bus_group[0].set_range((0.1, 0.2, 0.3, 0.4))
        control_bus_group[0].set_range((0.1, 0.2, 0.3, 0.4))
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/c_setn", 0, 4, 0.1, 0.2, 0.3, 0.4),
    ]



# flake8: noqa
import os
import platform

import pytest

from supriya.ugens import (
    DecodeB2,
    LFNoise2,
    Out,
    PanB2,
    PinkNoise,
    SuperColliderSynthDef,
    SynthDefBuilder,
)


@pytest.fixture
def py_synthdef():
    with SynthDefBuilder() as builder:
        source = PinkNoise.ar()
        azimuth = LFNoise2.kr(frequency=0.25)
        w, x, y = PanB2.ar(source=source, azimuth=azimuth)
        source = DecodeB2.ar(channel_count=4, w=w, x=x, y=y, orientation=0.5)
        Out.ar(bus=0, source=source)
    py_synthdef = builder.build("ambisonics")
    return py_synthdef


def test_ambisonics_supriya_vs_bytes(py_synthdef):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\nambisonics'
                b'\x00\x00\x00\x04'
                    b'>\x80\x00\x00'
                    b'?\x80\x00\x00'
                    b'?\x00\x00\x00'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x05'
                    b'\tPinkNoise'
                        b'\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x02'
                    b'\x08LFNoise2'
                        b'\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\x01'
                    b'\x05PanB2'
                        b'\x02'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x01'
                                b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x01'
                            b'\x02'
                            b'\x02'
                            b'\x02'
                    b'\x08DecodeB2'
                        b'\x02'
                        b'\x00\x00\x00\x04'
                        b'\x00\x00\x00\x04'
                        b'\x00\x00'
                            b'\x00\x00\x00\x02'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x02'
                                b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x02'
                                b'\x00\x00\x00\x02'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x02'
                            b'\x02'
                            b'\x02'
                            b'\x02'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x05'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x03'
                            b'\x00\x00\x00\x03'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x03'
                                b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x03'
                                b'\x00\x00\x00\x02'
                            b'\x00\x00\x00\x03'
                                b'\x00\x00\x00\x03'
                b'\x00\x00'
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_ambisonics_supriya_vs_sclang(py_synthdef):
    sc_synthdef = SuperColliderSynthDef(
        "ambisonics",
        r"""
        var source, azimuth, w, x, y;
        source = PinkNoise.ar();
        azimuth = LFNoise2.kr(0.25);
        #w, x, y = PanB2.ar(source, azimuth, 1);
        source = DecodeB2.ar(4, w, x, y, 0.5);
        Out.ar(0, source);
        """,
    )
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef.compile()
    assert sc_compiled_synthdef == py_compiled_synthdef



import pytest

from supriya.patterns import SequencePattern
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, sequence, iterations, expected, is_infinite",
    [
        (None, [1, 2, 3], None, [1, 2, 3], True),
        (None, [1, 2, 3], 1, [1, 2, 3], False),
        (None, [1, 2, 3], 2, [1, 2, 3, 1, 2, 3], False),
        (None, [1, 2, 3, SequencePattern(["a", "b"])], 1, [1, 2, 3, "a", "b"], False),
        (
            None,
            [1, 2, 3, SequencePattern(["a", "b"], None)],
            1,
            [1, 2, 3, "a", "b"],
            True,
        ),
        (
            None,
            [SequencePattern([1, 2, 3]), SequencePattern(["a", "b"])],
            1,
            [1, 2, 3, "a", "b"],
            False,
        ),
        (
            None,
            [SequencePattern([1, 2, 3]), SequencePattern(["a", "b"])],
            2,
            [1, 2, 3, "a", "b", 1, 2, 3, "a", "b"],
            False,
        ),
        (
            None,
            [SequencePattern([1, 2, 3], None), SequencePattern(["a", "b"])],
            1,
            [1, 2, 3],
            True,
        ),
        (
            None,
            [SequencePattern([1, 2, 3], None), SequencePattern(["a", "b"])],
            None,
            [1, 2, 3],
            True,
        ),
    ],
)
def test(stop_at, sequence, iterations, expected, is_infinite):
    pattern = SequencePattern(sequence, iterations=iterations)
    run_pattern_test(pattern, expected, is_infinite, stop_at)


@pytest.mark.parametrize(
    "sequence, iterations, raises",
    [
        ([1, 2, 3], 1, None),
        ([1, 2, 3], 10, None),
        ([1, 2, 3], None, None),
        ([1, 2, 3], 0, ValueError),
        (23, 1, ValueError),
    ],
)
def test___init__(sequence, iterations, raises):
    if raises:
        with pytest.raises(raises):
            SequencePattern(sequence, iterations)
    else:
        SequencePattern(sequence, iterations)



from .clap import clap
from .default import default
from .kick import kick
from .multiband_compressor import multiband_compressor
from .simple_sine import simple_sine
from .sweep_filter import sweep_filter
from .system_synthdefs import (
    system_link_audio_1,
    system_link_audio_2,
    system_link_audio_3,
    system_link_audio_4,
    system_link_audio_5,
    system_link_audio_6,
    system_link_audio_7,
    system_link_audio_8,
    system_link_audio_9,
    system_link_audio_10,
    system_link_audio_11,
    system_link_audio_12,
    system_link_audio_13,
    system_link_audio_14,
    system_link_audio_15,
    system_link_audio_16,
    system_link_control_1,
    system_link_control_2,
    system_link_control_3,
    system_link_control_4,
    system_link_control_5,
    system_link_control_6,
    system_link_control_7,
    system_link_control_8,
    system_link_control_9,
    system_link_control_10,
    system_link_control_11,
    system_link_control_12,
    system_link_control_13,
    system_link_control_14,
    system_link_control_15,
    system_link_control_16,
)
from .test import test, test_two_voice

__all__ = [
    "clap",
    "default",
    "kick",
    "multiband_compressor",
    "simple_sine",
    "sweep_filter",
    "system_link_audio_1",
    "system_link_audio_10",
    "system_link_audio_11",
    "system_link_audio_12",
    "system_link_audio_13",
    "system_link_audio_14",
    "system_link_audio_15",
    "system_link_audio_16",
    "system_link_audio_2",
    "system_link_audio_3",
    "system_link_audio_4",
    "system_link_audio_5",
    "system_link_audio_6",
    "system_link_audio_7",
    "system_link_audio_8",
    "system_link_audio_9",
    "system_link_control_1",
    "system_link_control_10",
    "system_link_control_11",
    "system_link_control_12",
    "system_link_control_13",
    "system_link_control_14",
    "system_link_control_15",
    "system_link_control_16",
    "system_link_control_2",
    "system_link_control_3",
    "system_link_control_4",
    "system_link_control_5",
    "system_link_control_6",
    "system_link_control_7",
    "system_link_control_8",
    "system_link_control_9",
    "test",
    "test_two_voice",
]



from typing import Tuple


def midi_note_number_to_frequency(midi_note_number):
    return pow(2.0, (midi_note_number - 69) / 12) * 440.0


def midi_velocity_to_amplitude(midi_velocity):
    return pow(midi_velocity / 127.0, 2.0)


def midi_velocity_to_decibels(midi_velocity):
    pass


def amplitude_to_decibels(amplitude):
    pass


def decibels_to_amplitude(decibels):
    pass


def measure_to_offset(
    measure: int,
    time_signature: Tuple[int, int],
    previous_measure: int,
    previous_time_signature_change_offset: float,
) -> float:
    return (
        (measure - previous_measure) * (time_signature[0] / time_signature[1])
    ) + previous_time_signature_change_offset


def offset_to_measure(
    offset: float,
    time_signature: Tuple[int, int],
    previous_measure: int,
    previous_time_signature_change_offset: float,
) -> int:
    return (
        int(
            (offset - previous_time_signature_change_offset)
            // (time_signature[0] / time_signature[1])
        )
        + previous_measure
    )


def offset_to_measure_offset(
    offset: float,
    time_signature: Tuple[int, int],
    previous_time_signature_change_offset: float,
) -> float:
    return (offset - previous_time_signature_change_offset) % (
        time_signature[0] / time_signature[1]
    )


def offset_to_seconds(
    beats_per_minute: float,
    current_offset: float,
    previous_offset: float,
    previous_seconds: float,
    beat_duration: float,
) -> float:
    return (
        (current_offset - previous_offset) / (beats_per_minute / 60) / beat_duration
    ) + previous_seconds


def seconds_to_offset(
    beats_per_minute: float,
    current_time: float,
    previous_offset: float,
    previous_seconds: float,
    beat_duration: float,
) -> float:
    return (
        (current_time - previous_seconds) * (beats_per_minute / 60) * beat_duration
    ) + previous_offset



from uqbar.strings import normalize

from supriya.ugens import Mix, Out, SinOsc, SynthDefBuilder


def test_multi_value_parameters():
    with SynthDefBuilder(amp=0.1, freqs=[300, 400], out=0) as builder:
        sines = SinOsc.ar(frequency=builder["freqs"])
        sines = Mix.new(sines)
        sines = sines * builder["amp"]
        Out.ar(bus=builder["out"], source=sines)
    synthdef = builder.build()
    assert str(synthdef) == normalize(
        """
        synthdef:
            name: 58528261cb129f5bee634d41a34e082c
            ugens:
            -   Control.kr:
                    amp: 0.1
                    freqs[0]: 300.0
                    freqs[1]: 400.0
                    out: 0.0
            -   SinOsc.ar/0:
                    frequency: Control.kr[1:freqs[0]]
                    phase: 0.0
            -   SinOsc.ar/1:
                    frequency: Control.kr[2:freqs[1]]
                    phase: 0.0
            -   BinaryOpUGen(ADDITION).ar:
                    left: SinOsc.ar/0[0]
                    right: SinOsc.ar/1[0]
            -   BinaryOpUGen(MULTIPLICATION).ar:
                    left: BinaryOpUGen(ADDITION).ar[0]
                    right: Control.kr[0:amp]
            -   Out.ar:
                    bus: Control.kr[3:out]
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
        """
    )



import pytest

import supriya.ugens
from supriya import SynthDefBuilder
from supriya.ugens.core import SynthDefError


def test_01():
    """
    Cannot share UGens.
    """

    with SynthDefBuilder():
        sine_one = supriya.ugens.SinOsc.ar()
        supriya.ugens.Out.ar(bus=0, source=sine_one)

    with SynthDefBuilder():
        sine_two = supriya.ugens.SinOsc.ar()
        with pytest.raises(SynthDefError) as exception_info:
            sine_two * sine_one
        assert "UGen input in different scope" in str(exception_info.value)


def test_02():
    """
    Cannot share parameters.
    """

    with SynthDefBuilder(bus=0) as builder:
        sine_one = supriya.ugens.SinOsc.ar()
        synth_one_bus = builder["bus"]
        supriya.ugens.Out.ar(bus=synth_one_bus, source=sine_one)

    with SynthDefBuilder():
        sine_two = supriya.ugens.SinOsc.ar()
        with pytest.raises(SynthDefError) as exception_info:
            supriya.ugens.Out.ar(bus=synth_one_bus, source=sine_two)
        assert "UGen input in different scope" in str(exception_info.value)


def test_03():
    """
    Cannot share output proxies.
    """
    with SynthDefBuilder():
        left, right = supriya.ugens.SinOsc.ar(frequency=[440, 442])
        supriya.ugens.Out.ar(bus=0, source=[right, left])

    with SynthDefBuilder():
        with pytest.raises(SynthDefError) as exception_info:
            left * right
        assert "UGen input in different scope" in str(exception_info.value)



from supriya.enums import EnvelopeShape
from supriya.ugens import (
    EnvGen,
    Envelope,
    In,
    InFeedback,
    Out,
    SynthDef,
    SynthDefBuilder,
)


def _build_link_audio_synthdef(channel_count: int) -> SynthDef:
    with SynthDefBuilder(
        out=0, in_=16, gate=1, fade_time=0.02, done_action=2
    ) as builder:
        start_value = builder["fade_time"] <= 0
        envelope = Envelope(
            amplitudes=[start_value, 1.0, 0.0],
            durations=[1.0, 1.0],
            curves=[EnvelopeShape.SINE],
            release_node=1,
        )
        envgen = EnvGen.kr(
            done_action=builder["done_action"],
            envelope=envelope,
            gate=builder["gate"],
            time_scale=builder["fade_time"],
        )
        source = InFeedback.ar(bus=builder["in_"], channel_count=channel_count)
        Out.ar(bus=builder["out"], source=source * envgen)
    return builder.build(name=f"system_link_audio_{channel_count}")


def _build_link_control_synthdef(channel_count: int) -> SynthDef:
    with SynthDefBuilder(
        out=0, in_=16, gate=1, fade_time=0.02, done_action=2
    ) as builder:
        start_value = builder["fade_time"] <= 0
        envelope = EnvGen.kr(
            done_action=builder["done_action"],
            envelope=Envelope(
                amplitudes=[start_value, 1.0, 0.0],
                durations=[1.0, 1.0],
                curves=[EnvelopeShape.LINEAR],
                release_node=1,
            ),
            gate=builder["gate"],
            time_scale=builder["fade_time"],
        )
        input_ = In.kr(bus=builder["in_"], channel_count=channel_count)
        Out.kr(bus=builder["out"], source=input_ * envelope)
    return builder.build(name=f"system_link_control_{channel_count}")


system_link_audio_1 = _build_link_audio_synthdef(1)
system_link_audio_2 = _build_link_audio_synthdef(2)
system_link_audio_3 = _build_link_audio_synthdef(3)
system_link_audio_4 = _build_link_audio_synthdef(4)
system_link_audio_5 = _build_link_audio_synthdef(5)
system_link_audio_6 = _build_link_audio_synthdef(6)
system_link_audio_7 = _build_link_audio_synthdef(7)
system_link_audio_8 = _build_link_audio_synthdef(8)
system_link_audio_9 = _build_link_audio_synthdef(9)
system_link_audio_10 = _build_link_audio_synthdef(10)
system_link_audio_11 = _build_link_audio_synthdef(11)
system_link_audio_12 = _build_link_audio_synthdef(12)
system_link_audio_13 = _build_link_audio_synthdef(13)
system_link_audio_14 = _build_link_audio_synthdef(14)
system_link_audio_15 = _build_link_audio_synthdef(15)
system_link_audio_16 = _build_link_audio_synthdef(16)
system_link_control_1 = _build_link_control_synthdef(1)
system_link_control_2 = _build_link_control_synthdef(2)
system_link_control_3 = _build_link_control_synthdef(3)
system_link_control_4 = _build_link_control_synthdef(4)
system_link_control_5 = _build_link_control_synthdef(5)
system_link_control_6 = _build_link_control_synthdef(6)
system_link_control_7 = _build_link_control_synthdef(7)
system_link_control_8 = _build_link_control_synthdef(8)
system_link_control_9 = _build_link_control_synthdef(9)
system_link_control_10 = _build_link_control_synthdef(10)
system_link_control_11 = _build_link_control_synthdef(11)
system_link_control_12 = _build_link_control_synthdef(12)
system_link_control_13 = _build_link_control_synthdef(13)
system_link_control_14 = _build_link_control_synthdef(14)
system_link_control_15 = _build_link_control_synthdef(15)
system_link_control_16 = _build_link_control_synthdef(16)

__all__ = [
    "system_link_audio_1",
    "system_link_audio_2",
    "system_link_audio_3",
    "system_link_audio_4",
    "system_link_audio_5",
    "system_link_audio_6",
    "system_link_audio_7",
    "system_link_audio_8",
    "system_link_audio_9",
    "system_link_audio_10",
    "system_link_audio_11",
    "system_link_audio_12",
    "system_link_audio_13",
    "system_link_audio_14",
    "system_link_audio_15",
    "system_link_audio_16",
    "system_link_control_1",
    "system_link_control_2",
    "system_link_control_3",
    "system_link_control_4",
    "system_link_control_5",
    "system_link_control_6",
    "system_link_control_7",
    "system_link_control_8",
    "system_link_control_9",
    "system_link_control_10",
    "system_link_control_11",
    "system_link_control_12",
    "system_link_control_13",
    "system_link_control_14",
    "system_link_control_15",
    "system_link_control_16",
]
